#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import asyncio
import threading
import queue
import unittest.mock
from concurrent import futures

import gel
from gel import _testbase as tb


class Barrier:
    def __init__(self, number):
        self._counter = number
        self._cond = threading.Condition()

    def ready(self):
        if self._counter == 0:
            return
        with self._cond:
            self._counter -= 1
            assert self._counter >= 0, self._counter
            if self._counter == 0:
                self._cond.notify_all()
            else:
                self._cond.wait_for(lambda: self._counter == 0)


class TestSyncRetry(tb.SyncQueryTestCase):

    SETUP = '''
        CREATE TYPE test::Counter EXTENDING std::Object {
            CREATE PROPERTY name -> std::str {
                CREATE CONSTRAINT std::exclusive;
            };
            CREATE PROPERTY value -> std::int32 {
                SET default := 0;
            };
        };
    '''

    TEARDOWN = '''
        DROP TYPE test::Counter;
    '''

    def test_sync_retry_01(self):
        for tx in self.client.transaction():
            with tx:
                tx.execute('''
                    INSERT test::Counter {
                        name := 'counter1'
                    };
                ''')

    def test_sync_retry_02(self):
        with self.assertRaises(ZeroDivisionError):
            for tx in self.client.transaction():
                with tx:
                    tx.execute('''
                        INSERT test::Counter {
                            name := 'counter_retry_02'
                        };
                    ''')
                    1 / 0
        with self.assertRaises(gel.NoDataError):
            self.client.query_required_single('''
                SELECT test::Counter
                FILTER .name = 'counter_retry_02'
            ''')
        self.assertEqual(
            self.client.query_single('''
                SELECT test::Counter
                FILTER .name = 'counter_retry_02'
            '''),
            None
        )

    def test_sync_retry_begin(self):
        patcher = unittest.mock.patch(
            "edgedb.base_client.BaseConnection.privileged_execute"
        )
        _start = patcher.start()

        def cleanup():
            try:
                patcher.stop()
            except RuntimeError:
                pass

        self.addCleanup(cleanup)

        _start.side_effect = gel.BackendUnavailableError()

        with self.assertRaises(gel.BackendUnavailableError):
            for tx in self.client.transaction():
                with tx:
                    tx.execute('''
                        INSERT test::Counter {
                            name := 'counter_retry_begin'
                        };
                    ''')
        with self.assertRaises(gel.NoDataError):
            self.client.query_required_single('''
                SELECT test::Counter
                FILTER .name = 'counter_retry_begin'
            ''')
        self.assertEqual(
            self.client.query_single('''
                SELECT test::Counter
                FILTER .name = 'counter_retry_begin'
            '''),
            None
        )

        def recover_after_first_error(*_, **__):
            patcher.stop()
            raise gel.BackendUnavailableError()

        _start.side_effect = recover_after_first_error
        call_count = _start.call_count

        for tx in self.client.transaction():
            with tx:
                tx.execute('''
                    INSERT test::Counter {
                        name := 'counter_retry_begin'
                    };
                ''')
        self.assertEqual(_start.call_count, call_count + 1)
        self.client.query_single('''
            SELECT test::Counter
            FILTER .name = 'counter_retry_begin'
        ''')

    def test_sync_retry_conflict(self):
        self.execute_conflict('counter2')

    def test_sync_conflict_no_retry(self):
        with self.assertRaises(gel.TransactionSerializationError):
            self.execute_conflict(
                'counter3',
                gel.RetryOptions(attempts=1, backoff=gel.default_backoff)
            )

    def execute_conflict(self, name='counter2', options=None):
        con_args = self.get_connect_args().copy()
        con_args.update(database=self.get_database_name())
        client2 = gel.create_client(**con_args)
        self.addCleanup(client2.close)

        barrier = Barrier(2)
        lock = threading.Lock()

        iterations = 0

        def transaction1(client):
            for tx in client.transaction():
                nonlocal iterations
                iterations += 1
                with tx:
                    # This magic query makes the test more reliable for some
                    # reason. I guess this is because starting a transaction
                    # in EdgeDB (and/or Postgres) is accomplished somewhat
                    # lazily, i.e. only start transaction on the first query
                    # rather than on the `START TRANSACTION`.
                    tx.query("SELECT 1")

                    # Start both transactions at the same initial data.
                    # One should succeed other should fail and retry.
                    # On next attempt, the latter should succeed
                    barrier.ready()

                    lock.acquire()
                    res = tx.query_single('''
                        SELECT (
                            INSERT test::Counter {
                                name := <str>$name,
                                value := 1,
                            } UNLESS CONFLICT ON .name
                            ELSE (
                                UPDATE test::Counter
                                SET { value := .value + 1 }
                            )
                        ).value
                    ''', name=name)
                lock.release()
            return res

        client = self.client
        if options:
            client = client.with_retry_options(options)
            client2 = client2.with_retry_options(options)

        with futures.ThreadPoolExecutor(2) as pool:
            f1 = pool.submit(transaction1, client)
            f2 = pool.submit(transaction1, client2)
            results = {f1.result(), f2.result()}

        self.assertEqual(results, {1, 2})
        self.assertEqual(iterations, 3)

    def test_sync_transaction_interface_errors(self):
        with self.assertRaisesRegex(
            AttributeError,
            "'Iteration' object has no attribute 'start'",
        ):
            for tx in self.client.transaction():
                with tx:
                    tx.start()

        with self.assertRaisesRegex(
            AttributeError,
            "'Iteration' object has no attribute 'rollback'",
        ):
            for tx in self.client.transaction():
                with tx:
                    tx.rollback()

        with self.assertRaisesRegex(
            AttributeError,
            "'Iteration' object has no attribute 'start'",
        ):
            for tx in self.client.transaction():
                tx.start()

        with self.assertRaisesRegex(gel.InterfaceError,
                                    r'.*Use `with transaction:`'):
            for tx in self.client.transaction():
                tx.execute("SELECT 123")

        with self.assertRaisesRegex(
            gel.InterfaceError,
            r"already in a `with` block",
        ):
            for tx in self.client.transaction():
                with tx:
                    with tx:
                        pass

    def test_sync_retry_parse(self):
        loop = asyncio.new_event_loop()
        q = queue.Queue()

        async def init():
            return asyncio.Event(), asyncio.Event()

        reconnect, terminate = loop.run_until_complete(init())

        async def proxy(r, w):
            try:
                while True:
                    buf = await r.read(65536)
                    if not buf:
                        w.close()
                        break
                    w.write(buf)
            except asyncio.CancelledError:
                pass

        async def cb(ri, wi):
            try:
                args = self.get_connect_args()
                ro, wo = await asyncio.open_connection(
                    args["host"], args["port"]
                )
                try:
                    fs = [
                        asyncio.create_task(proxy(ri, wo)),
                        asyncio.create_task(proxy(ro, wi)),
                        asyncio.create_task(terminate.wait()),
                    ]
                    if not reconnect.is_set():
                        fs.append(asyncio.create_task(reconnect.wait()))
                    _, pending = await asyncio.wait(
                        fs, return_when=asyncio.FIRST_COMPLETED
                    )
                    for f in pending:
                        f.cancel()
                finally:
                    wo.close()
            finally:
                wi.close()

        async def proxy_server():
            srv = await asyncio.start_server(cb, host="127.0.0.1", port=0)
            try:
                q.put(srv.sockets[0].getsockname()[1])
                await terminate.wait()
            finally:
                srv.close()
                await srv.wait_closed()

        with futures.ThreadPoolExecutor(1) as pool:
            pool.submit(loop.run_until_complete, proxy_server())
            try:
                client = self.make_test_client(
                    host="127.0.0.1",
                    port=q.get(),
                    database=self.get_database_name(),
                )

                # Fill the connection pool with a healthy connection
                self.assertEqual(client.query_single("SELECT 42"), 42)

                # Cut the connection to simulate an Internet interruption
                loop.call_soon_threadsafe(reconnect.set)

                # Run a new query that was never compiled, retry should work
                self.assertEqual(client.query_single("SELECT 1*2+3-4"), 1)
            finally:
                loop.call_soon_threadsafe(terminate.set)



import pathlib
import re
import warnings

from ..introspection import get_mod_and_name, GelORMWarning, FilePrinter


GEL_SCALAR_MAP = {
    'std::uuid': 'UUIDField',
    'std::bigint': 'DecimalField',
    'std::bool': 'BooleanField',
    'std::bytes': 'BinaryField',
    'std::decimal': 'DecimalField',
    'std::float32': 'FloatField',
    'std::float64': 'FloatField',
    'std::int16': 'SmallIntegerField',
    'std::int32': 'IntegerField',
    'std::int64': 'BigIntegerField',
    'std::json': 'JSONField',
    'std::str': 'TextField',
    # Extreme caution is needed for datetime field, the TZ aware and naive
    # values are controlled in Django via settings (USE_TZ) and are mutually
    # exclusive in the same app under default circumstances.
    'std::datetime': 'DateTimeField',
    'std::cal::local_date': 'DateField',
    'std::cal::local_datetime': 'DateTimeField',
    'std::cal::local_time': 'TimeField',
    # all kinds of durations are not supported due to this error:
    # iso_8601 intervalstyle currently not supported
}

BASE_STUB = f'''\
#
# Automatically generated from Gel schema.
#
# This is based on the auto-generated Django model module, which has been
# updated to fit Gel schema more closely.
#

from django.db import models
from django.contrib.postgres import fields as pgf


class GelUUIDField(models.UUIDField):
    # This field must be treated as a auto-generated UUID.
    db_returning = True


class LTForeignKey(models.ForeignKey):
    # Linked tables need to return their source/target ForeignKeys.
    db_returning = True\
'''

GEL_META = f'''
class GelPGMeta:
    'This is a model reflected from Gel using Postgres protocol.'
'''

CLOSEPAR_RE = re.compile(r'\)(?=\s+#|$)')
ARRAY_RE = re.compile(r'^array<(?P<el>.+)>$')
NAME_RE = re.compile(r'^(?P<alpha>\w+?)(?P<num>\d*)$')


def field_name_sort(item):
    key, val = item

    match = NAME_RE.fullmatch(key)
    res = (match.group('alpha'), int(match.group('num') or -1))

    return res


class ModelClass(object):
    def __init__(self, name):
        self.name = name
        self.props = {}
        self.links = {}
        self.mlinks = {}
        self.meta = {'managed': False}
        self.backlinks = {}
        self.backlink_renames = {}

    @property
    def table(self):
        return self.meta['db_table'].strip("'")

    def get_backlink_name(self, name, srcname):
        return f'_{name}_{srcname}'


class ModelGenerator(FilePrinter):
    def __init__(self, *, out):
        super().__init__()
        # record the output file path
        self.outfile = pathlib.Path(out).resolve()

    def spec_to_modules_dict(self, spec):
        modules = {
            mod: {'link_tables': {}, 'object_types': {}}
            for mod in sorted(spec['modules'])
        }

        for rec in spec['link_tables']:
            mod = rec['module']
            modules[mod]['link_tables'][rec['table']] = rec

        for rec in spec['object_types']:
            mod, name = get_mod_and_name(rec['name'])
            modules[mod]['object_types'][name] = rec

        return modules['default']

    def replace_foreignkey(self, fval, origtarget, newtarget, bkname=None):
        # Replace the reference with the string quoted
        # (because we don't check the order of definition)
        # name.
        fval = fval.replace(origtarget, repr(newtarget))

        if bkname:
            # Add a backlink reference
            fval = CLOSEPAR_RE.sub(f', related_name={bkname!r})', fval)

        return fval

    def build_models(self, maps):
        modmap = {}

        for name, rec in maps['object_types'].items():
            mod = ModelClass(name)
            mod.meta['db_table'] = repr(name)
            if 'backlink_renames' in rec:
                mod.backlink_renames = rec['backlink_renames']

            # copy backlink information
            for link in rec['backlinks']:
                mod.backlinks[link['name']] = link

            # process properties as fields
            for prop in rec['properties']:
                pname = prop['name']
                if pname == 'id' or prop['cardinality'] == 'Many':
                    continue

                code = self.render_prop(prop)
                if code:
                    mod.props[pname] = code

            # process single links as fields
            for link in rec['links']:
                if link['cardinality'] != 'One':
                    # Multi links require link tables and are handled
                    # separately.
                    continue

                lname = link['name']
                bklink = mod.get_backlink_name(lname, name)
                code = self.render_link(link, bklink)
                if code:
                    mod.links[lname] = code

            modmap[mod.name] = mod

        for table, rec in maps['link_tables'].items():
            source, fwname = table.split('.')
            mod = ModelClass(f'{source}{fwname.title()}')
            mod.meta['db_table'] = repr(table)
            mod.meta['unique_together'] = "(('source', 'target'),)"

            # Only have source and target
            mtgt, target = get_mod_and_name(rec['target'])
            if mtgt != 'default':
                # skip this whole link table
                warnings.warn(
                    f'Skipping link {fwname!r}: link target '
                    f'{rec["target"]!r} is not supported',
                    GelORMWarning,
                )
                continue

            mod.links['source'] = (
                f"LTForeignKey({source!r}, models.DO_NOTHING, "
                f"db_column='source')"
            )
            mod.links['target'] = (
                f"LTForeignKey({target!r}, models.DO_NOTHING, "
                f"db_column='target', primary_key=True)"
            )

            # Update the source model with the corresponding
            # ManyToManyField.
            src = modmap[source]
            tgt = modmap[target]
            bkname = src.get_backlink_name(fwname, source)
            src.mlinks[fwname] = (
                f'models.ManyToManyField('
                f'{tgt.name!r}, '
                f'through={mod.name!r}, '
                f'through_fields=("source", "target"), '
                f'related_name={bkname!r})'
            )

            modmap[mod.name] = mod

        return modmap

    def render_prop(self, prop):
        if prop['required']:
            req = ''
        else:
            req = 'blank=True, null=True'

        target = prop['target']['name']
        is_array = False
        match = ARRAY_RE.fullmatch(target)
        if match:
            is_array = True
            target = match.group('el')

        try:
            ftype = GEL_SCALAR_MAP[target]
        except KeyError:
            warnings.warn(
                f'Scalar type {target} is not supported',
                GelORMWarning,
            )
            return ''

        if is_array:
            return f'pgf.ArrayField(models.{ftype}({req}))'
        else:
            return f'models.{ftype}({req})'

    def render_link(self, link, bklink=None):
        if link['required']:
            req = ''
        else:
            req = ', blank=True, null=True'

        mod, target = get_mod_and_name(link['target']['name'])

        if mod != 'default':
            warnings.warn(
                f'Skipping link {link["name"]!r}: link target '
                f'{link["target"]["name"]!r} is not supported',
                GelORMWarning,
            )
            return ''

        if bklink:
            bklink = f', related_name={bklink!r}'
        else:
            bklink = ''

        return (f'models.ForeignKey('
                f'{target!r}, models.DO_NOTHING{bklink}{req})')

    def render_models(self, spec):
        # Check that there is only "default" module
        mods = spec['modules']
        if mods[0] != 'default' or len(mods) > 1:
            skipped = ', '.join([repr(m) for m in mods if m != 'default'])
            warnings.warn(
                f"Skipping modules {skipped}: Django reflection doesn't "
                f"support multiple modules or non-default modules.",
                GelORMWarning,
            )
        # Check that we don't have multiprops or link properties as they
        # produce models without `id` field and Django doesn't like that. It
        # causes Django to mistakenly use `source` as `id` and also attempt to
        # UPDATE `target` on link tables.
        if len(spec['prop_objects']) > 0:
            warnings.warn(
                f"Skipping multi properties: Django reflection doesn't "
                f"support multi properties as they produce models without "
                f"`id` field.",
                GelORMWarning,
            )
        if len(spec['link_objects']) > 0:
            warnings.warn(
                f"Skipping link properties: Django reflection doesn't support "
                f"link properties as they produce models without `id` field.",
                GelORMWarning,
            )

        maps = self.spec_to_modules_dict(spec)
        modmap = self.build_models(maps)

        with open(self.outfile, 'w+t') as f:
            self.out = f
            self.write(BASE_STUB)

            for mod in sorted(modmap.values(), key=lambda x: x.name):
                self.write()
                self.write()
                self.render_model_class(mod)

    def render_model_class(self, mod):
        self.write(f'class {mod.name}(models.Model):')
        self.indent()

        if '.' not in mod.table:
            # This is only valid for regular objects, not link tables.
            self.write(f"id = GelUUIDField(primary_key=True)")
            self.write(f"gel_type_id = models.UUIDField(db_column='__type__')")

        if mod.props:
            self.write()
            self.write(f'# properties as Fields')
            props = sorted(mod.props.items(), key=field_name_sort)
            for name, val in props:
                self.write(f'{name} = {val}')

        if mod.links:
            self.write()
            self.write(f'# links as ForeignKeys')
            links = sorted(mod.links.items(), key=field_name_sort)
            for name, val in links:
                self.write(f'{name} = {val}')

        if mod.mlinks:
            self.write()
            self.write(f'# multi links as ManyToManyFields')
            mlinks = sorted(mod.mlinks.items(), key=field_name_sort)
            for name, val in mlinks:
                self.write(f'{name} = {val}')

        if '.' not in mod.table:
            self.write(GEL_META)

        self.write('class Meta:')
        self.indent()
        for name, val in mod.meta.items():
            self.write(f'{name} = {val}')
        self.dedent()

        self.dedent()


#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import itertools
from concurrent.futures import ThreadPoolExecutor

import edgedb

from gel import _testbase as tb
from edgedb import TransactionOptions


class TestSyncTx(tb.SyncQueryTestCase):

    SETUP = '''
        CREATE TYPE test::TransactionTest EXTENDING std::Object {
            CREATE PROPERTY name -> std::str;
        };
    '''

    TEARDOWN = '''
        DROP TYPE test::TransactionTest;
    '''

    def test_sync_transaction_regular_01(self):
        tr = self.client.transaction()

        with self.assertRaises(ZeroDivisionError):
            for with_tr in tr:
                with with_tr:
                    with_tr.execute('''
                        INSERT test::TransactionTest {
                            name := 'Test Transaction'
                        };
                    ''')

                    1 / 0

        result = self.client.query('''
            SELECT
                test::TransactionTest
            FILTER
                test::TransactionTest.name = 'Test Transaction';
        ''')

        self.assertEqual(result, [])

    async def test_sync_transaction_kinds(self):
        isolations = [
            None,
            edgedb.IsolationLevel.Serializable,
        ]
        booleans = [None, True, False]
        all = itertools.product(isolations, booleans, booleans)
        for isolation, readonly, deferrable in all:
            opt = dict(
                isolation=isolation,
                readonly=readonly,
                deferrable=deferrable,
            )
            # skip None
            opt = {k: v for k, v in opt.items() if v is not None}
            client = self.client.with_transaction_options(
                TransactionOptions(**opt)
            )
            try:
                for tx in client.transaction():
                    with tx:
                        tx.execute(
                            'INSERT test::TransactionTest {name := "test"}')
            except edgedb.TransactionError:
                self.assertTrue(readonly)
            else:
                self.assertFalse(readonly)

            for tx in client.transaction():
                with tx:
                    pass

    def test_sync_transaction_commit_failure(self):
        with self.assertRaises(edgedb.errors.QueryError):
            for tx in self.client.transaction():
                with tx:
                    tx.execute("start migration to {};")
        self.assertEqual(self.client.query_single("select 42"), 42)

    def test_sync_transaction_exclusive(self):
        for tx in self.client.transaction():
            with tx:
                query = "select sys::_sleep(0.5)"
                with ThreadPoolExecutor(max_workers=2) as executor:
                    f1 = executor.submit(tx.execute, query)
                    f2 = executor.submit(tx.execute, query)
                    with self.assertRaisesRegex(
                        edgedb.InterfaceError,
                        "concurrent queries within the same transaction "
                        "are not allowed"
                    ):
                        f1.result(timeout=5)
                        f2.result(timeout=5)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import dataclasses
import gc
import os
import random
import string
import unittest
import weakref

import edgedb
from edgedb.datatypes import datatypes as private
from edgedb import introspect


class TestRecordDesc(unittest.TestCase):

    def test_recorddesc_1(self):
        with self.assertRaisesRegex(TypeError, 'one to three positional'):
            private._RecordDescriptor()

        with self.assertRaisesRegex(TypeError, 'one to three positional'):
            private._RecordDescriptor(t=1)

        with self.assertRaisesRegex(TypeError, 'requires a tuple'):
            private._RecordDescriptor(1)

        with self.assertRaisesRegex(TypeError, 'requires a tuple'):
            private._RecordDescriptor(('a',), 1)

        with self.assertRaisesRegex(TypeError,
                                    'the same length as the names tuple'):
            private._RecordDescriptor(('a',), ())

        private._RecordDescriptor(('a', 'b'))

        with self.assertRaisesRegex(ValueError, f'more than {0x4000 - 1}'):
            private._RecordDescriptor(('a',) * 20000)

    def test_recorddesc_2(self):
        rd = private._RecordDescriptor(
            ('a', 'b', 'c'),
            (private._EDGE_POINTER_IS_LINKPROP,
             0,
             private._EDGE_POINTER_IS_LINK))

        self.assertEqual(rd.get_pos('a'), 0)
        self.assertEqual(rd.get_pos('b'), 1)
        self.assertEqual(rd.get_pos('c'), 2)

        self.assertTrue(rd.is_linkprop('a'))
        self.assertFalse(rd.is_linkprop('b'))
        self.assertFalse(rd.is_linkprop('c'))

        self.assertFalse(rd.is_link('a'))
        self.assertFalse(rd.is_link('b'))
        self.assertTrue(rd.is_link('c'))

        with self.assertRaises(LookupError):
            rd.get_pos('z')

        with self.assertRaises(LookupError):
            rd.is_linkprop('z')

    def test_recorddesc_3(self):
        f = private.create_object_factory(
            id={'property', 'implicit'},
            lb='link-property',
            c='property',
            d='link',
        )

        o = f(1, 2, 3, 4)

        desc = private.get_object_descriptor(o)
        self.assertEqual(set(dir(desc)), set(('id', '@lb', 'c', 'd')))

        self.assertTrue(desc.is_linkprop('@lb'))
        self.assertFalse(desc.is_linkprop('id'))
        self.assertFalse(desc.is_linkprop('c'))
        self.assertFalse(desc.is_linkprop('d'))

        self.assertFalse(desc.is_link('@lb'))
        self.assertFalse(desc.is_link('id'))
        self.assertFalse(desc.is_link('c'))
        self.assertTrue(desc.is_link('d'))

        self.assertFalse(desc.is_implicit('@lb'))
        self.assertTrue(desc.is_implicit('id'))
        self.assertFalse(desc.is_implicit('c'))
        self.assertFalse(desc.is_implicit('d'))

        self.assertEqual(desc.get_pos('@lb'), 1)
        self.assertEqual(desc.get_pos('id'), 0)
        self.assertEqual(desc.get_pos('c'), 2)
        self.assertEqual(desc.get_pos('d'), 3)

    def test_recorddesc_4(self):
        f = private.create_object_factory(
            id={'property', 'implicit'},
            lb='link-property',
            c='property',
            d='link',
        )

        o = f(1, 2, 3, 4)
        intro = introspect.introspect_object(o)

        self.assertEqual(
            intro.pointers,
            (
                ('id', introspect.ElementKind.PROPERTY, True),
                ('c', introspect.ElementKind.PROPERTY, False),
                ('d', introspect.ElementKind.LINK, False),
            )
        )

        # clear cache so that tests in refcount mode don't freak out.
        introspect._introspect_object_desc.cache_clear()


class TestTuple(unittest.TestCase):

    def test_tuple_empty_1(self):
        t = edgedb.Tuple()
        self.assertIsInstance(t, tuple)
        self.assertEqual(len(t), 0)
        self.assertEqual(hash(t), hash(()))
        self.assertEqual(repr(t), '()')
        with self.assertRaisesRegex(IndexError, 'out of range'):
            t[0]

    def test_tuple_2(self):
        t = edgedb.Tuple((1, 'a'))
        self.assertEqual(len(t), 2)
        self.assertEqual(hash(t), hash((1, 'a')))

        self.assertEqual(repr(t), "(1, 'a')")

        self.assertEqual(t[0], 1)
        self.assertEqual(t[1], 'a')
        with self.assertRaisesRegex(IndexError, 'out of range'):
            t[2]

    def test_tuple_3(self):
        t = edgedb.Tuple((1, []))
        t[1].append(t)
        self.assertEqual(t[1], [t])

        self.assertEqual(repr(t), '(1, [(...)])')
        self.assertEqual(str(t), '(1, [(...)])')

    def test_tuple_freelist_1(self):
        lst = []
        for _ in range(5000):
            lst.append(edgedb.Tuple((1,)))
        for t in lst:
            self.assertEqual(t[0], 1)

    def test_tuple_5(self):
        self.assertEqual(
            edgedb.Tuple([1, 2, 3]),
            edgedb.Tuple([1, 2, 3]))

        self.assertNotEqual(
            edgedb.Tuple([1, 2, 3]),
            edgedb.Tuple([1, 3, 2]))

        self.assertLess(
            edgedb.Tuple([1, 2, 3]),
            edgedb.Tuple([1, 3, 2]))

        self.assertEqual(
            edgedb.Tuple([]),
            edgedb.Tuple([]))

        self.assertEqual(
            edgedb.Tuple([1]),
            edgedb.Tuple([1]))

        self.assertGreaterEqual(
            edgedb.Tuple([1]),
            edgedb.Tuple([1]))

        self.assertNotEqual(
            edgedb.Tuple([1]),
            edgedb.Tuple([]))

        self.assertGreater(
            edgedb.Tuple([1]),
            edgedb.Tuple([]))

        self.assertNotEqual(
            edgedb.Tuple([1]),
            edgedb.Tuple([2]))

        self.assertLess(
            edgedb.Tuple([1]),
            edgedb.Tuple([2]))

        self.assertNotEqual(
            edgedb.Tuple([1, 2]),
            edgedb.Tuple([2, 2]))

        self.assertNotEqual(
            edgedb.Tuple([1, 1]),
            edgedb.Tuple([2, 2, 1]))

    def test_tuple_6(self):
        self.assertEqual(
            edgedb.Tuple([1, 2, 3]),
            (1, 2, 3))

        self.assertEqual(
            (1, 2, 3),
            edgedb.Tuple([1, 2, 3]))

        self.assertNotEqual(
            edgedb.Tuple([1, 2, 3]),
            (1, 3, 2))

        self.assertLess(
            edgedb.Tuple([1, 2, 3]),
            (1, 3, 2))

        self.assertEqual(
            edgedb.Tuple([]),
            ())

        self.assertEqual(
            edgedb.Tuple([1]),
            (1,))

        self.assertGreaterEqual(
            edgedb.Tuple([1]),
            (1,))

        self.assertNotEqual(
            edgedb.Tuple([1]),
            ())

        self.assertGreater(
            edgedb.Tuple([1]),
            ())

        self.assertNotEqual(
            edgedb.Tuple([1]),
            (2,))

        self.assertLess(
            edgedb.Tuple([1]),
            (2,))

        self.assertNotEqual(
            edgedb.Tuple([1, 2]),
            (2, 2))

        self.assertNotEqual(
            edgedb.Tuple([1, 1]),
            (2, 2, 1))

    def test_tuple_7(self):
        self.assertNotEqual(
            edgedb.Tuple([1, 2, 3]),
            123)


class TestNamedTuple(unittest.TestCase):

    def test_namedtuple_empty_1(self):
        with self.assertRaisesRegex(ValueError, 'at least one field'):
            edgedb.NamedTuple()

    def test_namedtuple_2(self):
        t = edgedb.NamedTuple(a=1)
        self.assertEqual(repr(t), "(a := 1)")

        t = edgedb.NamedTuple(a=1, b='a')

        self.assertEqual(set(dir(t)), {'a', 'b'})

        self.assertEqual(repr(t), "(a := 1, b := 'a')")

        self.assertEqual(t[0], 1)
        self.assertEqual(t[1], 'a')
        with self.assertRaisesRegex(IndexError, 'out of range'):
            t[2]

        self.assertEqual(len(t), 2)
        self.assertEqual(hash(t), hash((1, 'a')))

        self.assertEqual(t.a, 1)
        self.assertEqual(t.b, 'a')

        with self.assertRaises(AttributeError):
            t.z

    def test_namedtuple_3(self):
        t = edgedb.NamedTuple(a=1, b=[])
        t.b.append(t)
        self.assertEqual(t.b, [t])

        self.assertEqual(repr(t), '(a := 1, b := [(...)])')
        self.assertEqual(str(t), '(a := 1, b := [(...)])')

    def test_namedtuple_4(self):
        t1 = edgedb.NamedTuple(a=1, b='aaaa')
        t2 = edgedb.Tuple((1, 'aaaa'))
        t3 = (1, 'aaaa')

        self.assertEqual(hash(t1), hash(t2))
        self.assertEqual(hash(t1), hash(t3))

    def test_namedtuple_5(self):
        self.assertEqual(
            edgedb.NamedTuple(a=1, b=2, c=3),
            edgedb.NamedTuple(x=1, y=2, z=3))

        self.assertNotEqual(
            edgedb.NamedTuple(a=1, b=2, c=3),
            edgedb.NamedTuple(a=1, c=3, b=2))

        self.assertLess(
            edgedb.NamedTuple(a=1, b=2, c=3),
            edgedb.NamedTuple(a=1, b=3, c=2))

        self.assertEqual(
            edgedb.NamedTuple(a=1),
            edgedb.NamedTuple(b=1))

        self.assertEqual(
            edgedb.NamedTuple(a=1),
            edgedb.NamedTuple(a=1))

    def test_namedtuple_6(self):
        self.assertEqual(
            edgedb.NamedTuple(a=1, b=2, c=3),
            (1, 2, 3))

        self.assertEqual(
            (1, 2, 3),
            edgedb.NamedTuple(a=1, b=2, c=3))

        self.assertNotEqual(
            edgedb.NamedTuple(a=1, b=2, c=3),
            (1, 3, 2))

        self.assertLess(
            edgedb.NamedTuple(a=1, b=2, c=3),
            (1, 3, 2))

        self.assertEqual(
            edgedb.NamedTuple(a=1),
            (1,))

        self.assertEqual(
            edgedb.NamedTuple(a=1),
            (1,))

    def test_namedtuple_7(self):
        self.assertNotEqual(
            edgedb.NamedTuple(a=1, b=2, c=3),
            1)

    def test_namedtuple_8(self):
        self.assertEqual(
            edgedb.NamedTuple(壹=1, 贰=2, 叄=3),
            (1, 2, 3))

    def test_namedtuple_memory(self):
        num = int(os.getenv("EDGEDB_PYTHON_TEST_NAMEDTUPLE_MEMORY", 100))

        def test():
            nt = []
            fix_tp = type(edgedb.NamedTuple(a=1, b=2))
            for _i in range(num):
                values = {}
                for _ in range(random.randint(9, 16)):
                    key = "".join(random.choices(string.ascii_letters, k=3))
                    value = random.randint(16384, 65536)
                    values[key] = value
                nt.append(edgedb.NamedTuple(**values))
                if random.random() > 0.5:
                    nt.append(
                        fix_tp(random.randint(10, 20), random.randint(20, 30))
                    )
                if len(nt) % random.randint(10, 20) == 0:
                    nt[:] = nt[random.randint(5, len(nt)):]

        gc.collect()
        gc.collect()
        gc.collect()
        gc_count = gc.get_count()
        test()
        gc.collect()
        gc.collect()
        gc.collect()
        self.assertEqual(gc.get_count(), gc_count)


class TestDerivedNamedTuple(unittest.TestCase):
    DerivedNamedTuple = type(edgedb.NamedTuple(a=1, b=2, c=3))

    def test_derived_namedtuple_1(self):
        self.assertEqual(
            (1, 2, 3),
            self.DerivedNamedTuple(a=1, b=2, c=3),
        )
        self.assertEqual(
            (1, 2, 3),
            self.DerivedNamedTuple(c=3, b=2, a=1),
        )
        self.assertEqual(
            (1, 2, 3),
            self.DerivedNamedTuple(1, c=3, b=2),
        )
        self.assertEqual(
            (1, 2, 3),
            self.DerivedNamedTuple(1, 2, 3),
        )

    def test_derived_namedtuple_2(self):
        with self.assertRaisesRegex(ValueError, "requires 3 arguments"):
            self.DerivedNamedTuple()

        with self.assertRaisesRegex(ValueError, "requires 3 arguments"):
            self.DerivedNamedTuple(1)

        with self.assertRaisesRegex(ValueError, "only needs 3 arguments"):
            self.DerivedNamedTuple(1, 2, 3, 4)

    def test_derived_namedtuple_3(self):
        with self.assertRaisesRegex(ValueError, "missing required argument"):
            self.DerivedNamedTuple(a=1)

        with self.assertRaisesRegex(ValueError, "missing required argument"):
            self.DerivedNamedTuple(b=2)

        with self.assertRaisesRegex(ValueError, "missing required argument"):
            self.DerivedNamedTuple(1, 2, d=4)

        with self.assertRaisesRegex(ValueError, "extra keyword arguments"):
            self.DerivedNamedTuple(1, 2, 3, d=4)

        with self.assertRaisesRegex(ValueError, "extra keyword arguments"):
            self.DerivedNamedTuple(1, 2, c=3, d=4)

    def test_derived_namedtuple_4(self):
        tp = type(edgedb.NamedTuple(x=42))
        tp(8)
        edgedb.NamedTuple(y=88)
        tp(16)
        tp_ref = weakref.ref(tp)
        gc.collect()
        self.assertIsNotNone(tp_ref())
        del tp
        gc.collect()
        self.assertIsNone(tp_ref())


class TestObject(unittest.TestCase):

    def test_object_1(self):
        f = private.create_object_factory(
            id='property',
            lb='link-property',
            c='property'
        )

        o = f(1, 2, 3)

        self.assertEqual(repr(o), 'Object{id := 1, @lb := 2, c := 3}')

        self.assertEqual(o.id, 1)
        self.assertEqual(o.c, 3)

        with self.assertRaises(AttributeError):
            o.lb

        with self.assertRaises(AttributeError):
            o.z

        with self.assertRaises(TypeError):
            len(o)

        with self.assertRaises(TypeError):
            o[0]

        with self.assertRaises(TypeError):
            o['id']

        self.assertEqual(set(dir(o)), {'id', 'c'})

    def test_object_2(self):
        f = private.create_object_factory(
            id={'property', 'implicit'},
            lb='link-property',
            c='property'
        )

        o = f(1, 2, 3)

        self.assertEqual(repr(o), 'Object{@lb := 2, c := 3}')

        self.assertNotEqual(hash(o), hash(f(1, 2, 3)))
        self.assertNotEqual(hash(o), hash(f(1, 2, 'aaaa')))
        self.assertNotEqual(hash(o), hash((1, 2, 3)))

        self.assertEqual(set(dir(o)), {'id', 'c'})

    def test_object_3(self):
        f = private.create_object_factory(id='property', c='link')
        o = f(1, [])

        o.c.append(o)
        self.assertEqual(repr(o), 'Object{id := 1, c := [Object{...}]}')

    def test_object_4(self):
        f = private.create_object_factory(
            id={'property', 'implicit'},
            lb='link-property',
            c='property'
        )

        o1 = f(1, 'aa', 'ba')
        o2 = f(1, 'ab', 'bb')
        o3 = f(3, 'ac', 'bc')

        self.assertNotEqual(o1, o2)
        self.assertNotEqual(o1, o3)

    def test_object_5(self):
        f = private.create_object_factory(
            a='property',
            lb='link-property',
            c='property'
        )
        x = f(1, 2, 3)
        self.assertFalse(hasattr(x, 'id'))

    def test_object_6(self):
        User = private.create_object_factory(
            id='property',
            name='property',
        )

        u = User(1, 'user1')

        with self.assertRaisesRegex(TypeError,
                                    "property 'name' should be "
                                    "accessed via dot notation"):
            u['name']

    def test_object_links_1(self):
        O2 = private.create_object_factory(
            id='property',
            lb='link-property',
            c='property'
        )

        O1 = private.create_object_factory(
            id='property',
            o2s='link'
        )

        o2_1 = O2(1, 'linkprop o2 1', 3)
        o2_2 = O2(4, 'linkprop o2 2', 6)
        o1 = O1(2, edgedb.Set((o2_1, o2_2)))

        with self.assertRaisesRegex(TypeError,
                                    "link 'o2s' should be "
                                    "accessed via dot notation"):
            o1['o2s']

    def test_object_link_property_1(self):
        O2 = private.create_object_factory(
            id='property',
            lb='link-property',
            c='property'
        )

        O1 = private.create_object_factory(
            id='property',
            o2s='link'
        )

        o2_1 = O2(1, 'linkprop o2 1', 3)
        o2_2 = O2(4, 'linkprop o2 2', 6)
        o1 = O1(2, edgedb.Set((o2_1, o2_2)))

        o2s = o1.o2s
        self.assertEqual(len(o2s), 2)
        self.assertEqual(o2s, o1.o2s)
        self.assertEqual(
            repr(o2s),
            "[Object{id := 1, @lb := 'linkprop o2 1', c := 3},"
            " Object{id := 4, @lb := 'linkprop o2 2', c := 6}]"
        )

        self.assertEqual(o2s[0]['@lb'], 'linkprop o2 1')
        self.assertEqual(o2s[1]['@lb'], 'linkprop o2 2')
        self.assertEqual(getattr(o2s[0], '@lb'), 'linkprop o2 1')
        self.assertEqual(getattr(o2s[1], '@lb'), 'linkprop o2 2')

        with self.assertRaises(AttributeError):
            o2s[0].lb

        with self.assertRaises(AttributeError):
            getattr(o2s[0], "@lb2")

        with self.assertRaisesRegex(
            TypeError,
            "link property 'lb' should be accessed with '@' prefix",
        ):
            o2s[0]['lb']

        with self.assertRaisesRegex(
            TypeError, "property 'c' should be accessed via dot notation"
        ):
            o2s[0]['c']

        with self.assertRaisesRegex(
            KeyError, "link property '@c' does not exist"
        ):
            o2s[0]['@c']

    def test_object_dataclass_1(self):
        User = private.create_object_factory(
            id='property',
            name='property',
            tuple='property',
            namedtuple='property',
            linkprop="link-property",
        )

        u = User(
            1,
            'Bob',
            edgedb.Tuple((1, 2.0, '3')),
            edgedb.NamedTuple(a=1, b="Y"),
            123,
        )
        self.assertTrue(dataclasses.is_dataclass(u))
        self.assertEqual(
            dataclasses.asdict(u),
            {
                'id': 1,
                'name': 'Bob',
                'tuple': (1, 2.0, '3'),
                'namedtuple': (1, "Y"),
            },
        )


class TestSet(unittest.TestCase):

    def test_set_1(self):
        s = edgedb.Set(())
        self.assertEqual(repr(s), '[]')

        s = edgedb.Set((1, 2, [], 'a'))

        self.assertEqual(s[1], 2)
        self.assertEqual(s[2], [])
        self.assertEqual(len(s), 4)
        with self.assertRaises(IndexError):
            s[10]

    def test_set_2(self):
        s = edgedb.Set((1, 2, 3000, 'a'))

        self.assertEqual(repr(s), "[1, 2, 3000, 'a']")

    def test_set_3(self):
        s = edgedb.Set(())

        self.assertEqual(len(s), 0)

    def test_set_4(self):
        s = edgedb.Set(([],))
        s[0].append(s)
        self.assertEqual(repr(s), "[[[...]]]")

    def test_set_5(self):
        self.assertNotEqual(
            edgedb.Set([1, 2, 3]),
            edgedb.Set([3, 2, 1]))

        self.assertEqual(
            edgedb.Set([]),
            edgedb.Set([]))

        self.assertEqual(
            edgedb.Set([1]),
            edgedb.Set([1]))

        self.assertNotEqual(
            edgedb.Set([1]),
            edgedb.Set([]))

        self.assertNotEqual(
            edgedb.Set([1]),
            edgedb.Set([2]))

        self.assertNotEqual(
            edgedb.Set([1, 2]),
            edgedb.Set([2, 2]))

        self.assertNotEqual(
            edgedb.Set([1, 1, 2]),
            edgedb.Set([2, 2, 1]))

    def test_set_6(self):
        f = private.create_object_factory(
            id={'property', 'implicit'},
            lb='link-property',
            c='property'
        )

        o1 = f(1, 'aa', edgedb.Set([1, 2, 3]))
        o2 = f(1, 'ab', edgedb.Set([1, 2, 4]))
        o3 = f(3, 'ac', edgedb.Set([5, 5, 5, 5]))

        self.assertNotEqual(
            edgedb.Set([o1, o2, o3]),
            edgedb.Set([o2, o3, o1]))

        self.assertNotEqual(
            edgedb.Set([o1, o3]),
            edgedb.Set([o2, o3]))

        self.assertNotEqual(
            edgedb.Set([o1, o1]),
            edgedb.Set([o2, o3]))

    def test_set_7(self):
        self.assertEqual(
            edgedb.Set([1, 2, 3]),
            [1, 2, 3])

        self.assertNotEqual(
            edgedb.Set([1, 2, 3]),
            [3, 2, 1])

        self.assertNotEqual(
            edgedb.Set([1, 2, 3]),
            1)

    def test_set_8(self):
        s = edgedb.Set([1, 2, 3])
        si = iter(s)
        self.assertEqual(list(si), [1, 2, 3])


class TestArray(unittest.TestCase):

    def test_array_empty_1(self):
        t = edgedb.Array()
        self.assertEqual(len(t), 0)
        with self.assertRaisesRegex(IndexError, 'out of range'):
            t[0]
        self.assertEqual(repr(t), "[]")

    def test_array_2(self):
        t = edgedb.Array((1, 'a'))

        self.assertEqual(repr(t), "[1, 'a']")
        self.assertEqual(str(t), "[1, 'a']")

        self.assertEqual(len(t), 2)

        self.assertEqual(t[0], 1)
        self.assertEqual(t[1], 'a')
        with self.assertRaisesRegex(IndexError, 'out of range'):
            t[2]

    def test_array_3(self):
        t = edgedb.Array((1, []))
        t[1].append(t)
        self.assertEqual(t[1], [t])
        self.assertEqual(repr(t), '[1, [[...]]]')

    def test_array_4(self):
        self.assertEqual(
            edgedb.Array([1, 2, 3]),
            edgedb.Array([1, 2, 3]))

        self.assertNotEqual(
            edgedb.Array([1, 2, 3]),
            edgedb.Array([1, 3, 2]))

        self.assertLess(
            edgedb.Array([1, 2, 3]),
            edgedb.Array([1, 3, 2]))

        self.assertEqual(
            edgedb.Array([]),
            edgedb.Array([]))

        self.assertEqual(
            edgedb.Array([1]),
            edgedb.Array([1]))

        self.assertGreaterEqual(
            edgedb.Array([1]),
            edgedb.Array([1]))

        self.assertNotEqual(
            edgedb.Array([1]),
            edgedb.Array([]))

        self.assertGreater(
            edgedb.Array([1]),
            edgedb.Array([]))

        self.assertNotEqual(
            edgedb.Array([1]),
            edgedb.Array([2]))

        self.assertLess(
            edgedb.Array([1]),
            edgedb.Array([2]))

        self.assertNotEqual(
            edgedb.Array([1, 2]),
            edgedb.Array([2, 2]))

        self.assertNotEqual(
            edgedb.Array([1, 1]),
            edgedb.Array([2, 2, 1]))

    def test_array_5(self):
        self.assertEqual(
            edgedb.Array([1, 2, 3]),
            [1, 2, 3])

        self.assertEqual(
            [1, 2, 3],
            edgedb.Array([1, 2, 3]))

        self.assertNotEqual(
            [1, 2, 4],
            edgedb.Array([1, 2, 3]))

        self.assertNotEqual(
            edgedb.Array([1, 2, 3]),
            [1, 3, 2])

        self.assertLess(
            edgedb.Array([1, 2, 3]),
            [1, 3, 2])

        self.assertEqual(
            edgedb.Array([]),
            [])

        self.assertEqual(
            edgedb.Array([1]),
            [1])

        self.assertGreaterEqual(
            edgedb.Array([1]),
            [1])

        self.assertNotEqual(
            edgedb.Array([1]),
            [])

        self.assertGreater(
            edgedb.Array([1]),
            [])

        self.assertNotEqual(
            edgedb.Array([1]),
            [2])

        self.assertLess(
            edgedb.Array([1]),
            [2])

        self.assertNotEqual(
            edgedb.Array([1, 2]),
            [2, 2])

        self.assertNotEqual(
            edgedb.Array([1, 1]),
            [2, 2, 1])

    def test_array_6(self):
        self.assertNotEqual(
            edgedb.Array([1, 2, 3]),
            False)


class TestRange(unittest.TestCase):

    def test_range_empty_1(self):
        t = edgedb.Range(empty=True)
        self.assertEqual(t.lower, None)
        self.assertEqual(t.upper, None)
        self.assertFalse(t.inc_lower)
        self.assertFalse(t.inc_upper)
        self.assertTrue(t.is_empty())
        self.assertFalse(t)

        self.assertEqual(t, edgedb.Range(1, 1, empty=True))

        with self.assertRaisesRegex(ValueError, 'conflicting arguments'):
            edgedb.Range(1, 2, empty=True)

    def test_range_2(self):
        t = edgedb.Range(1, 2)
        self.assertEqual(repr(t), "<Range [1, 2]>")
        self.assertEqual(str(t), "<Range [1, 2]>")

        self.assertEqual(t.lower, 1)
        self.assertEqual(t.upper, 2)
        self.assertTrue(t.inc_lower)
        self.assertFalse(t.inc_upper)
        self.assertFalse(t.is_empty())
        self.assertTrue(t)

    def test_range_3(self):
        t = edgedb.Range(1)
        self.assertEqual(t.lower, 1)
        self.assertEqual(t.upper, None)
        self.assertTrue(t.inc_lower)
        self.assertFalse(t.inc_upper)
        self.assertFalse(t.is_empty())

        t = edgedb.Range(None, 1)
        self.assertEqual(t.lower, None)
        self.assertEqual(t.upper, 1)
        self.assertFalse(t.inc_lower)
        self.assertFalse(t.inc_upper)
        self.assertFalse(t.is_empty())

        t = edgedb.Range(None, None)
        self.assertEqual(t.lower, None)
        self.assertEqual(t.upper, None)
        self.assertFalse(t.inc_lower)
        self.assertFalse(t.inc_upper)
        self.assertFalse(t.is_empty())

    def test_range_4(self):
        for il in (False, True):
            for iu in (False, True):
                t = edgedb.Range(1, 2, inc_lower=il, inc_upper=iu)
                self.assertEqual(t.lower, 1)
                self.assertEqual(t.upper, 2)
                self.assertEqual(t.inc_lower, il)
                self.assertEqual(t.inc_upper, iu)
                self.assertFalse(t.is_empty())

    def test_range_5(self):
        # test hash
        self.assertEqual(
            {
                edgedb.Range(None, 2, inc_upper=True),
                edgedb.Range(1, 2),
                edgedb.Range(1, 2),
                edgedb.Range(1, 2),
                edgedb.Range(None, 2, inc_upper=True),
            },
            {
                edgedb.Range(1, 2),
                edgedb.Range(None, 2, inc_upper=True),
            }
        )


class TestMultiRange(unittest.TestCase):

    def test_multirange_empty_1(self):
        t = edgedb.MultiRange()
        self.assertEqual(len(t), 0)
        self.assertEqual(t, edgedb.MultiRange([]))

    def test_multirange_2(self):
        t = edgedb.MultiRange([
            edgedb.Range(1, 2),
            edgedb.Range(4),
        ])
        self.assertEqual(
            repr(t), "<MultiRange [<Range [1, 2]>, <Range [4, ]>]>")
        self.assertEqual(
            str(t), "<MultiRange [<Range [1, 2]>, <Range [4, ]>]>")

        self.assertEqual(
            t,
            edgedb.MultiRange([
                edgedb.Range(1, 2),
                edgedb.Range(4),
            ])
        )

    def test_multirange_3(self):
        ranges = [
            edgedb.Range(None, 0),
            edgedb.Range(1, 2),
            edgedb.Range(4),
        ]
        t = edgedb.MultiRange([
            edgedb.Range(None, 0),
            edgedb.Range(1, 2),
            edgedb.Range(4),
        ])

        for el, r in zip(t, ranges):
            self.assertEqual(el, r)

    def test_multirange_4(self):
        # test hash
        self.assertEqual(
            {
                edgedb.MultiRange([
                    edgedb.Range(1, 2),
                    edgedb.Range(4),
                ]),
                edgedb.MultiRange([edgedb.Range(None, 2, inc_upper=True)]),
                edgedb.MultiRange([
                    edgedb.Range(1, 2),
                    edgedb.Range(4),
                ]),
                edgedb.MultiRange([
                    edgedb.Range(1, 2),
                    edgedb.Range(4),
                ]),
                edgedb.MultiRange([edgedb.Range(None, 2, inc_upper=True)]),
            },
            {
                edgedb.MultiRange([edgedb.Range(None, 2, inc_upper=True)]),
                edgedb.MultiRange([
                    edgedb.Range(1, 2),
                    edgedb.Range(4),
                ]),
            }
        )

    def test_multirange_5(self):
        # test hash
        self.assertEqual(
            edgedb.MultiRange([
                edgedb.Range(None, 2, inc_upper=True),
                edgedb.Range(5, 9),
                edgedb.Range(5, 9),
                edgedb.Range(5, 9),
                edgedb.Range(None, 2, inc_upper=True),
            ]),
            edgedb.MultiRange([
                edgedb.Range(5, 9),
                edgedb.Range(None, 2, inc_upper=True),
            ]),
        )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from collections import namedtuple, UserDict

import edgedb
from gel import _testbase as tb


class TestNamedTupleTypes(tb.SyncQueryTestCase):

    async def test_namedtuple_01(self):
        NT1 = namedtuple('NT2', ['x', 'y'])
        NT2 = namedtuple('NT2', ['y', 'x'])

        ctors = [dict, UserDict, NT1, NT2]
        for ctor in ctors:
            val = ctor(x=10, y='y')
            res = self.client.query_single('''
                select <tuple<x: int64, y: str>>$0
            ''', val)

            self.assertEqual(res, (10, 'y'))

    async def test_namedtuple_02(self):
        NT1 = namedtuple('NT2', ['x', 'z'])

        with self.assertRaisesRegex(edgedb.InvalidArgumentError, 'is missing'):
            self.client.query_single('''
                select <tuple<x: int64, y: str>>$0
            ''', dict(x=20, z='test'))

        with self.assertRaisesRegex(edgedb.InvalidArgumentError, 'is missing'):
            self.client.query_single('''
                select <tuple<x: int64, y: str>>$0
            ''', NT1(x=20, z='test'))



import os
import sys
import warnings

COLOR = None


class Color:
    HEADER = ""
    BLUE = ""
    CYAN = ""
    GREEN = ""
    WARNING = ""
    FAIL = ""
    ENDC = ""
    BOLD = ""
    UNDERLINE = ""


def get_color() -> Color:
    global COLOR

    if COLOR is None:
        COLOR = Color()
        if type(USE_COLOR) is bool:
            use_color = USE_COLOR
        else:
            try:
                use_color = USE_COLOR()
            except Exception:
                use_color = False
        if use_color:
            COLOR.HEADER = '\033[95m'
            COLOR.BLUE = '\033[94m'
            COLOR.CYAN = '\033[96m'
            COLOR.GREEN = '\033[92m'
            COLOR.WARNING = '\033[93m'
            COLOR.FAIL = '\033[91m'
            COLOR.ENDC = '\033[0m'
            COLOR.BOLD = '\033[1m'
            COLOR.UNDERLINE = '\033[4m'

    return COLOR


try:
    USE_COLOR = {
        "default": lambda: sys.stderr.isatty(),
        "auto": lambda: sys.stderr.isatty(),
        "enabled": True,
        "disabled": False,
    }[
        os.getenv("EDGEDB_COLOR_OUTPUT", "default")
    ]
except KeyError:
    warnings.warn(
        "EDGEDB_COLOR_OUTPUT can only be one of: "
        "default, auto, enabled or disabled",
        stacklevel=1,
    )
    USE_COLOR = False



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio
import itertools

import edgedb

from gel import _testbase as tb
from edgedb import TransactionOptions
from edgedb.options import RetryOptions


class TestAsyncTx(tb.AsyncQueryTestCase):

    SETUP = '''
        CREATE TYPE test::TransactionTest EXTENDING std::Object {
            CREATE PROPERTY name -> std::str;
        };
    '''

    TEARDOWN = '''
        DROP TYPE test::TransactionTest;
    '''

    async def test_async_transaction_regular_01(self):
        tr = self.client.with_retry_options(
            RetryOptions(attempts=1)).transaction()

        with self.assertRaises(ZeroDivisionError):
            async for with_tr in tr:
                async with with_tr:
                    await with_tr.execute('''
                        INSERT test::TransactionTest {
                            name := 'Test Transaction'
                        };
                    ''')

                    1 / 0

        result = await self.client.query('''
            SELECT
                test::TransactionTest
            FILTER
                test::TransactionTest.name = 'Test Transaction';
        ''')

        self.assertEqual(result, [])

    async def test_async_transaction_kinds(self):
        isolations = [
            None,
            edgedb.IsolationLevel.Serializable,
        ]
        booleans = [None, True, False]
        all = itertools.product(isolations, booleans, booleans)
        for isolation, readonly, deferrable in all:
            opt = dict(
                isolation=isolation,
                readonly=readonly,
                deferrable=deferrable,
            )
            # skip None
            opt = {k: v for k, v in opt.items() if v is not None}
            client = self.client.with_transaction_options(
                TransactionOptions(**opt)
            )
            async for tx in client.transaction():
                async with tx:
                    pass

    async def test_async_transaction_commit_failure(self):
        with self.assertRaises(edgedb.errors.QueryError):
            async for tx in self.client.transaction():
                async with tx:
                    await tx.execute("start migration to {};")
        self.assertEqual(await self.client.query_single("select 42"), 42)

    async def test_async_transaction_exclusive(self):
        async for tx in self.client.transaction():
            async with tx:
                query = "select sys::_sleep(0.01)"
                f1 = self.loop.create_task(tx.execute(query))
                f2 = self.loop.create_task(tx.execute(query))
                with self.assertRaisesRegex(
                    edgedb.InterfaceError,
                    "concurrent queries within the same transaction "
                    "are not allowed"
                ):
                    await asyncio.wait_for(f1, timeout=5)
                    await asyncio.wait_for(f2, timeout=5)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import abc
import random
import time
import typing

from . import abstract
from . import con_utils
from . import enums
from . import errors
from . import options as _options
from .protocol import protocol


BaseConnection_T = typing.TypeVar('BaseConnection_T', bound='BaseConnection')
QUERY_CACHE_SIZE = 1000


class BaseConnection(metaclass=abc.ABCMeta):
    _protocol: typing.Any
    _addr: typing.Optional[typing.Union[str, typing.Tuple[str, int]]]
    _addrs: typing.Iterable[typing.Union[str, typing.Tuple[str, int]]]
    _config: con_utils.ClientConfiguration
    _params: con_utils.ResolvedConnectConfig
    _log_listeners: typing.Set[
        typing.Callable[[BaseConnection_T, errors.EdgeDBMessage], None]
    ]
    __slots__ = (
        "__weakref__",
        "_protocol",
        "_addr",
        "_addrs",
        "_config",
        "_params",
        "_log_listeners",
        "_holder",
    )

    def __init__(
        self,
        addrs: typing.Iterable[typing.Union[str, typing.Tuple[str, int]]],
        config: con_utils.ClientConfiguration,
        params: con_utils.ResolvedConnectConfig,
    ):
        self._addr = None
        self._protocol = None
        self._addrs = addrs
        self._config = config
        self._params = params
        self._log_listeners = set()
        self._holder = None

    @abc.abstractmethod
    def _dispatch_log_message(self, msg):
        ...

    def _on_log_message(self, msg):
        if self._log_listeners:
            self._dispatch_log_message(msg)

    def connected_addr(self):
        return self._addr

    def _get_last_status(self) -> typing.Optional[str]:
        if self._protocol is None:
            return None
        status = self._protocol.last_status
        if status is not None:
            status = status.decode()
        return status

    def _cleanup(self):
        self._log_listeners.clear()
        if self._holder:
            self._holder._release_on_close()
            self._holder = None

    def add_log_listener(
        self: BaseConnection_T,
        callback: typing.Callable[[BaseConnection_T, errors.EdgeDBMessage],
                                  None]
    ) -> None:
        """Add a listener for EdgeDB log messages.

        :param callable callback:
            A callable receiving the following arguments:
            **connection**: a Connection the callback is registered with;
            **message**: the `gel.EdgeDBMessage` message.
        """
        self._log_listeners.add(callback)

    def remove_log_listener(
        self: BaseConnection_T,
        callback: typing.Callable[[BaseConnection_T, errors.EdgeDBMessage],
                                  None]
    ) -> None:
        """Remove a listening callback for log messages."""
        self._log_listeners.discard(callback)

    @property
    def dbname(self) -> str:
        return self._params.database

    @property
    def branch(self) -> str:
        return self._params.branch

    @abc.abstractmethod
    def is_closed(self) -> bool:
        ...

    @abc.abstractmethod
    async def connect_addr(self, addr, timeout):
        ...

    @abc.abstractmethod
    async def sleep(self, seconds):
        ...

    async def connect(self, *, single_attempt=False):
        start = time.monotonic()
        if single_attempt:
            max_time = 0
        else:
            max_time = start + self._config.wait_until_available
        iteration = 1

        while True:
            for addr in self._addrs:
                try:
                    await self.connect_addr(addr, self._config.connect_timeout)
                except TimeoutError as e:
                    if iteration == 1 or time.monotonic() < max_time:
                        continue
                    else:
                        raise errors.ClientConnectionTimeoutError(
                            f"connecting to {addr} failed in"
                            f" {self._config.connect_timeout} sec"
                        ) from e
                except errors.ClientConnectionError as e:
                    if (
                        e.has_tag(errors.SHOULD_RECONNECT) and
                        (iteration == 1 or time.monotonic() < max_time)
                    ):
                        continue
                    nice_err = e.__class__(
                        con_utils.render_client_no_connection_error(
                            e,
                            addr,
                            attempts=iteration,
                            duration=time.monotonic() - start,
                        ))
                    raise nice_err from e.__cause__
                else:
                    return

            iteration += 1
            await self.sleep(0.01 + random.random() * 0.2)

    async def privileged_execute(
        self, execute_context: abstract.ExecuteContext
    ):
        if self._protocol.is_legacy:
            await self._protocol.legacy_simple_query(
                execute_context.query.query, enums.Capability.ALL
            )
        else:
            await self._protocol.execute(
                execute_context.lower(allow_capabilities=enums.Capability.ALL)
            )

    def is_in_transaction(self) -> bool:
        """Return True if Connection is currently inside a transaction.

        :return bool: True if inside transaction, False otherwise.
        """
        return self._protocol.is_in_transaction()

    def get_settings(self) -> typing.Dict[str, typing.Any]:
        return self._protocol.get_settings()

    async def _retry_operation(self, func, retry_options, ctx):
        reconnect = False
        i = 0
        while True:
            i += 1
            try:
                if reconnect:
                    await self.connect(single_attempt=True)
                return await func()

            except errors.EdgeDBError as e:
                if retry_options is None:
                    raise
                if not e.has_tag(errors.SHOULD_RETRY):
                    raise e
                # A query is read-only if it has no capabilities i.e.
                # capabilities == 0. Read-only queries are safe to retry.
                # Explicit transaction conflicts as well.
                if (
                    ctx.capabilities != 0
                    and not isinstance(e, errors.TransactionConflictError)
                ):
                    raise e
                rule = retry_options.get_rule_for_exception(e)
                if i >= rule.attempts:
                    raise e
                await self.sleep(rule.backoff(i))
                reconnect = self.is_closed()

    async def raw_query(self, query_context: abstract.QueryContext):
        if self.is_closed():
            await self.connect()

        reconnect = False
        i = 0
        if self._protocol.is_legacy:
            allow_capabilities = enums.Capability.LEGACY_EXECUTE
        else:
            allow_capabilities = enums.Capability.EXECUTE
        ctx = query_context.lower(allow_capabilities=allow_capabilities)

        async def _inner():
            if self._protocol.is_legacy:
                return await self._protocol.legacy_execute_anonymous(ctx)
            else:
                res = await self._protocol.query(ctx)
                if ctx.warnings:
                    res = query_context.warning_handler(ctx.warnings, res)
                return res

        return await self._retry_operation(
            _inner, query_context.retry_options, ctx
        )

    async def _execute(self, execute_context: abstract.ExecuteContext) -> None:
        if self._protocol.is_legacy:
            if execute_context.query.args or execute_context.query.kwargs:
                raise errors.InterfaceError(
                    "Legacy protocol doesn't support arguments in execute()"
                )
            await self._protocol.legacy_simple_query(
                execute_context.query.query, enums.Capability.LEGACY_EXECUTE
            )
        else:
            ctx = execute_context.lower(
                allow_capabilities=enums.Capability.EXECUTE
            )
            async def _inner():
                res = await self._protocol.execute(ctx)
                if ctx.warnings:
                    res = execute_context.warning_handler(ctx.warnings, res)

            return await self._retry_operation(
                _inner, execute_context.retry_options, ctx
            )

    async def describe(
        self, describe_context: abstract.DescribeContext
    ) -> abstract.DescribeResult:
        ctx = describe_context.lower(
            allow_capabilities=enums.Capability.EXECUTE
        )
        await self._protocol._parse(ctx)
        return abstract.DescribeResult(
            input_type=ctx.in_dc.make_type(describe_context),
            output_type=ctx.out_dc.make_type(describe_context),
            output_cardinality=enums.Cardinality(ctx.cardinality[0]),
            capabilities=ctx.capabilities,
        )

    def terminate(self):
        if not self.is_closed():
            try:
                self._protocol.abort()
            finally:
                self._cleanup()

    def __repr__(self):
        if self.is_closed():
            return '<{classname} [closed] {id:#x}>'.format(
                classname=self.__class__.__name__, id=id(self))
        else:
            return '<{classname} [connected to {addr}] {id:#x}>'.format(
                classname=self.__class__.__name__,
                addr=self.connected_addr(),
                id=id(self))


class PoolConnectionHolder(abc.ABC):
    __slots__ = (
        "_con",
        "_pool",
        "_release_event",
        "_timeout",
        "_generation",
    )
    _event_class = NotImplemented

    def __init__(self, pool):

        self._pool = pool
        self._con = None

        self._timeout = None
        self._generation = None

        self._release_event = self._event_class()
        self._release_event.set()

    @abc.abstractmethod
    async def close(self, *, wait=True):
        ...

    @abc.abstractmethod
    async def wait_until_released(self, timeout=None):
        ...

    async def connect(self):
        if self._con is not None:
            raise errors.InternalClientError(
                'PoolConnectionHolder.connect() called while another '
                'connection already exists')

        self._con = await self._pool._get_new_connection()
        assert self._con._holder is None
        self._con._holder = self
        self._generation = self._pool._generation

    async def acquire(self) -> BaseConnection:
        if self._con is None or self._con.is_closed():
            self._con = None
            await self.connect()

        elif self._generation != self._pool._generation:
            # Connections have been expired, re-connect the holder.
            self._con._holder = None  # don't release the connection
            await self.close(wait=False)
            self._con = None
            await self.connect()

        self._release_event.clear()

        return self._con

    async def release(self, timeout):
        if self._release_event.is_set():
            raise errors.InternalClientError(
                'PoolConnectionHolder.release() called on '
                'a free connection holder')

        if self._con.is_closed():
            # This is usually the case when the connection is broken rather
            # than closed by the user, so we need to call _release_on_close()
            # here to release the holder back to the queue, because
            # self._con._cleanup() was never called. On the other hand, it is
            # safe to call self._release() twice - the second call is no-op.
            self._release_on_close()
            return

        self._timeout = None

        if self._generation != self._pool._generation:
            # The connection has expired because it belongs to
            # an older generation (BasePoolImpl.expire_connections() has
            # been called.)
            await self.close()
            return

        # Free this connection holder and invalidate the
        # connection proxy.
        self._release()

    def terminate(self):
        if self._con is not None:
            # AsyncIOConnection.terminate() will call _release_on_close() to
            # finish holder cleanup.
            self._con.terminate()

    def _release_on_close(self):
        self._release()
        self._con = None

    def _release(self):
        """Release this connection holder."""
        if self._release_event.is_set():
            # The holder is not checked out.
            return

        self._release_event.set()

        # Put ourselves back to the pool queue.
        self._pool._queue.put_nowait(self)


class BasePoolImpl(abc.ABC):
    __slots__ = (
        "_connect_args",
        "_codecs_registry",
        "_query_cache",
        "_connection_factory",
        "_queue",
        "_user_max_concurrency",
        "_max_concurrency",
        "_first_connect_lock",
        "_working_addr",
        "_working_config",
        "_working_params",
        "_holders",
        "_initialized",
        "_initializing",
        "_closing",
        "_closed",
        "_generation",
    )

    _holder_class = NotImplemented

    def __init__(
        self,
        connect_args,
        connection_factory,
        *,
        max_concurrency: typing.Optional[int],
    ):
        self._connection_factory = connection_factory
        self._connect_args = connect_args
        self._codecs_registry = protocol.CodecsRegistry()
        self._query_cache = protocol.LRUMapping(maxsize=QUERY_CACHE_SIZE)

        if max_concurrency is not None and max_concurrency <= 0:
            raise ValueError(
                'max_concurrency is expected to be greater than zero'
            )

        self._user_max_concurrency = max_concurrency
        self._max_concurrency = max_concurrency if max_concurrency else 1

        self._holders = []
        self._queue = None

        self._first_connect_lock = None
        self._working_addr = None
        self._working_config = None
        self._working_params = None

        self._closing = False
        self._closed = False
        self._generation = 0

    @abc.abstractmethod
    def _ensure_initialized(self):
        ...

    @abc.abstractmethod
    def _set_queue_maxsize(self, maxsize):
        ...

    @abc.abstractmethod
    async def _maybe_get_first_connection(self):
        ...

    @abc.abstractmethod
    async def acquire(self, timeout=None):
        ...

    @abc.abstractmethod
    async def _release(self, connection):
        ...

    @property
    def codecs_registry(self):
        return self._codecs_registry

    @property
    def query_cache(self):
        return self._query_cache

    def _resize_holder_pool(self):
        resize_diff = self._max_concurrency - len(self._holders)

        if (resize_diff > 0):
            if self._queue.maxsize != self._max_concurrency:
                self._set_queue_maxsize(self._max_concurrency)

            for _ in range(resize_diff):
                ch = self._holder_class(self)

                self._holders.append(ch)
                self._queue.put_nowait(ch)
        elif resize_diff < 0:
            # TODO: shrink the pool
            pass

    def get_max_concurrency(self):
        return self._max_concurrency

    def get_free_size(self):
        if self._queue is None:
            # Queue has not been initialized yet
            return self._max_concurrency

        return self._queue.qsize()

    def set_connect_args(self, dsn=None, **connect_kwargs):
        r"""Set the new connection arguments for this pool.

        The new connection arguments will be used for all subsequent
        new connection attempts.  Existing connections will remain until
        they expire. Use BasePoolImpl.expire_connections() to expedite
        the connection expiry.

        :param str dsn:
            Connection arguments specified using as a single string in
            the following format:
            ``gel://user:pass@host:port/database?option=value``.

        :param \*\*connect_kwargs:
            Keyword arguments for the
            :func:`~gel.asyncio_client.create_async_client` function.
        """

        connect_kwargs["dsn"] = dsn
        self._connect_args = connect_kwargs
        self._codecs_registry = protocol.CodecsRegistry()
        self._query_cache = protocol.LRUMapping(maxsize=QUERY_CACHE_SIZE)
        self._working_addr = None
        self._working_config = None
        self._working_params = None

    async def _get_first_connection(self):
        # First connection attempt on this pool.
        connect_config, client_config = con_utils.parse_connect_arguments(
            **self._connect_args,
            # ToDos
            command_timeout=None,
            server_settings=None,
        )
        con = self._connection_factory(
            [connect_config.address], client_config, connect_config
        )
        await con.connect()
        self._working_addr = con.connected_addr()
        self._working_config = client_config
        self._working_params = connect_config

        if self._user_max_concurrency is None:
            suggested_concurrency = con.get_settings().get(
                'suggested_pool_concurrency')
            if suggested_concurrency:
                self._max_concurrency = suggested_concurrency
                self._resize_holder_pool()
        return con

    async def _get_new_connection(self):
        con = None
        if self._working_addr is None:
            con = await self._maybe_get_first_connection()
        if con is None:
            assert self._working_addr is not None
            # We've connected before and have a resolved address,
            # and parsed options and config.
            con = self._connection_factory(
                [self._working_addr],
                self._working_config,
                self._working_params,
            )
            await con.connect()

        return con

    async def release(self, connection):

        if not isinstance(connection, BaseConnection):
            raise errors.InterfaceError(
                f'BasePoolImpl.release() received invalid connection: '
                f'{connection!r} does not belong to any connection pool'
            )

        ch = connection._holder
        if ch is None:
            # Already released, do nothing.
            return

        if ch._pool is not self:
            raise errors.InterfaceError(
                f'BasePoolImpl.release() received invalid connection: '
                f'{connection!r} is not a member of this pool'
            )

        return await self._release(ch)

    def terminate(self):
        """Terminate all connections in the pool."""
        if self._closed:
            return
        for ch in self._holders:
            ch.terminate()
        self._closed = True

    def expire_connections(self):
        """Expire all currently open connections.

        Cause all currently open connections to get replaced on the
        next query.
        """
        self._generation += 1

    async def ensure_connected(self):
        self._ensure_initialized()

        for ch in self._holders:
            if ch._con is not None and not ch._con.is_closed():
                return

        ch = self._holders[0]
        ch._con = None
        await ch.connect()


class BaseClient(abstract.BaseReadOnlyExecutor, _options._OptionsMixin):
    __slots__ = ("_impl", "_options")
    _impl_class = NotImplemented

    def __init__(
        self,
        *,
        connection_class,
        max_concurrency: typing.Optional[int],
        dsn=None,
        host: str = None,
        port: int = None,
        credentials: str = None,
        credentials_file: str = None,
        user: str = None,
        password: str = None,
        secret_key: str = None,
        database: str = None,
        branch: str = None,
        tls_ca: str = None,
        tls_ca_file: str = None,
        tls_security: str = None,
        tls_server_name: str = None,
        wait_until_available: int = 30,
        timeout: int = 10,
        **kwargs,
    ):
        super().__init__()
        connect_args = {
            "dsn": dsn,
            "host": host,
            "port": port,
            "credentials": credentials,
            "credentials_file": credentials_file,
            "user": user,
            "password": password,
            "secret_key": secret_key,
            "database": database,
            "branch": branch,
            "timeout": timeout,
            "tls_ca": tls_ca,
            "tls_ca_file": tls_ca_file,
            "tls_security": tls_security,
            "tls_server_name": tls_server_name,
            "wait_until_available": wait_until_available,
        }

        self._impl = self._impl_class(
            connect_args,
            connection_class=connection_class,
            max_concurrency=max_concurrency,
            **kwargs,
        )

    def _shallow_clone(self):
        new_client = self.__class__.__new__(self.__class__)
        new_client._impl = self._impl
        return new_client

    def _get_query_cache(self) -> abstract.QueryCache:
        return abstract.QueryCache(
            codecs_registry=self._impl.codecs_registry,
            query_cache=self._impl.query_cache,
        )

    def _get_retry_options(self) -> typing.Optional[_options.RetryOptions]:
        return self._options.retry_options

    def _get_state(self) -> _options.State:
        return self._options.state

    def _get_warning_handler(self) -> _options.WarningHandler:
        return self._options.warning_handler

    def _get_annotations(self) -> typing.Dict[str, str]:
        return self._options.annotations

    @property
    def max_concurrency(self) -> int:
        """Max number of connections in the pool."""

        return self._impl.get_max_concurrency()

    @property
    def free_size(self) -> int:
        """Number of available connections in the pool."""

        return self._impl.get_free_size()

    async def _query(self, query_context: abstract.QueryContext):
        con = await self._impl.acquire()
        try:
            return await con.raw_query(query_context)
        finally:
            await self._impl.release(con)

    async def _execute(self, execute_context: abstract.ExecuteContext) -> None:
        con = await self._impl.acquire()
        try:
            await con._execute(execute_context)
        finally:
            await self._impl.release(con)

    async def _describe(
        self, describe_context: abstract.DescribeContext
    ) -> abstract.DescribeResult:
        con = await self._impl.acquire()
        try:
            return await con.describe(describe_context)
        finally:
            await self._impl.release(con)

    def terminate(self):
        """Terminate all connections in the pool."""
        self._impl.terminate()



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


# flake8: noqa

from ._version import __version__

from gel.datatypes.datatypes import (
    Tuple, NamedTuple, EnumValue, RelativeDuration, DateDuration, ConfigMemory
)
from gel.datatypes.datatypes import Record, Set, Object, Array
from gel.datatypes.range import Range, MultiRange

from .abstract import (
    Executor, AsyncIOExecutor, ReadOnlyExecutor, AsyncIOReadOnlyExecutor,
)

from .asyncio_client import (
    create_async_client,
    AsyncIOClient
)

from .blocking_client import create_client, Client
from .enums import Cardinality, ElementKind
from .options import RetryCondition, IsolationLevel, default_backoff
from .options import RetryOptions, TransactionOptions
from .options import State

from .errors._base import EdgeDBError, EdgeDBMessage

__all__ = [
    "Array",
    "AsyncIOClient",
    "AsyncIOExecutor",
    "AsyncIOReadOnlyExecutor",
    "Cardinality",
    "Client",
    "ConfigMemory",
    "DateDuration",
    "EdgeDBError",
    "EdgeDBMessage",
    "ElementKind",
    "EnumValue",
    "Executor",
    "IsolationLevel",
    "MultiRange",
    "NamedTuple",
    "Object",
    "Range",
    "ReadOnlyExecutor",
    "RelativeDuration",
    "RetryCondition",
    "RetryOptions",
    "Set",
    "Record",
    "State",
    "TransactionOptions",
    "Tuple",
    "create_async_client",
    "create_client",
    "default_backoff",
]


# The below is generated by `make gen-errors`.
# DO NOT MODIFY BY HAND.
#
# <ERRORS-AUTOGEN>
from .errors import (
    InternalServerError,
    UnsupportedFeatureError,
    ProtocolError,
    BinaryProtocolError,
    UnsupportedProtocolVersionError,
    TypeSpecNotFoundError,
    UnexpectedMessageError,
    InputDataError,
    ParameterTypeMismatchError,
    StateMismatchError,
    ResultCardinalityMismatchError,
    CapabilityError,
    UnsupportedCapabilityError,
    DisabledCapabilityError,
    QueryError,
    InvalidSyntaxError,
    EdgeQLSyntaxError,
    SchemaSyntaxError,
    GraphQLSyntaxError,
    InvalidTypeError,
    InvalidTargetError,
    InvalidLinkTargetError,
    InvalidPropertyTargetError,
    InvalidReferenceError,
    UnknownModuleError,
    UnknownLinkError,
    UnknownPropertyError,
    UnknownUserError,
    UnknownDatabaseError,
    UnknownParameterError,
    SchemaError,
    SchemaDefinitionError,
    InvalidDefinitionError,
    InvalidModuleDefinitionError,
    InvalidLinkDefinitionError,
    InvalidPropertyDefinitionError,
    InvalidUserDefinitionError,
    InvalidDatabaseDefinitionError,
    InvalidOperatorDefinitionError,
    InvalidAliasDefinitionError,
    InvalidFunctionDefinitionError,
    InvalidConstraintDefinitionError,
    InvalidCastDefinitionError,
    DuplicateDefinitionError,
    DuplicateModuleDefinitionError,
    DuplicateLinkDefinitionError,
    DuplicatePropertyDefinitionError,
    DuplicateUserDefinitionError,
    DuplicateDatabaseDefinitionError,
    DuplicateOperatorDefinitionError,
    DuplicateViewDefinitionError,
    DuplicateFunctionDefinitionError,
    DuplicateConstraintDefinitionError,
    DuplicateCastDefinitionError,
    DuplicateMigrationError,
    SessionTimeoutError,
    IdleSessionTimeoutError,
    QueryTimeoutError,
    TransactionTimeoutError,
    IdleTransactionTimeoutError,
    ExecutionError,
    InvalidValueError,
    DivisionByZeroError,
    NumericOutOfRangeError,
    AccessPolicyError,
    QueryAssertionError,
    IntegrityError,
    ConstraintViolationError,
    CardinalityViolationError,
    MissingRequiredError,
    TransactionError,
    TransactionConflictError,
    TransactionSerializationError,
    TransactionDeadlockError,
    WatchError,
    ConfigurationError,
    AccessError,
    AuthenticationError,
    AvailabilityError,
    BackendUnavailableError,
    ServerOfflineError,
    BackendError,
    UnsupportedBackendFeatureError,
    LogMessage,
    WarningMessage,
    ClientError,
    ClientConnectionError,
    ClientConnectionFailedError,
    ClientConnectionFailedTemporarilyError,
    ClientConnectionTimeoutError,
    ClientConnectionClosedError,
    InterfaceError,
    QueryArgumentError,
    MissingArgumentError,
    UnknownArgumentError,
    InvalidArgumentError,
    NoDataError,
    InternalClientError,
)

__all__.extend([
    "InternalServerError",
    "UnsupportedFeatureError",
    "ProtocolError",
    "BinaryProtocolError",
    "UnsupportedProtocolVersionError",
    "TypeSpecNotFoundError",
    "UnexpectedMessageError",
    "InputDataError",
    "ParameterTypeMismatchError",
    "StateMismatchError",
    "ResultCardinalityMismatchError",
    "CapabilityError",
    "UnsupportedCapabilityError",
    "DisabledCapabilityError",
    "QueryError",
    "InvalidSyntaxError",
    "EdgeQLSyntaxError",
    "SchemaSyntaxError",
    "GraphQLSyntaxError",
    "InvalidTypeError",
    "InvalidTargetError",
    "InvalidLinkTargetError",
    "InvalidPropertyTargetError",
    "InvalidReferenceError",
    "UnknownModuleError",
    "UnknownLinkError",
    "UnknownPropertyError",
    "UnknownUserError",
    "UnknownDatabaseError",
    "UnknownParameterError",
    "SchemaError",
    "SchemaDefinitionError",
    "InvalidDefinitionError",
    "InvalidModuleDefinitionError",
    "InvalidLinkDefinitionError",
    "InvalidPropertyDefinitionError",
    "InvalidUserDefinitionError",
    "InvalidDatabaseDefinitionError",
    "InvalidOperatorDefinitionError",
    "InvalidAliasDefinitionError",
    "InvalidFunctionDefinitionError",
    "InvalidConstraintDefinitionError",
    "InvalidCastDefinitionError",
    "DuplicateDefinitionError",
    "DuplicateModuleDefinitionError",
    "DuplicateLinkDefinitionError",
    "DuplicatePropertyDefinitionError",
    "DuplicateUserDefinitionError",
    "DuplicateDatabaseDefinitionError",
    "DuplicateOperatorDefinitionError",
    "DuplicateViewDefinitionError",
    "DuplicateFunctionDefinitionError",
    "DuplicateConstraintDefinitionError",
    "DuplicateCastDefinitionError",
    "DuplicateMigrationError",
    "SessionTimeoutError",
    "IdleSessionTimeoutError",
    "QueryTimeoutError",
    "TransactionTimeoutError",
    "IdleTransactionTimeoutError",
    "ExecutionError",
    "InvalidValueError",
    "DivisionByZeroError",
    "NumericOutOfRangeError",
    "AccessPolicyError",
    "QueryAssertionError",
    "IntegrityError",
    "ConstraintViolationError",
    "CardinalityViolationError",
    "MissingRequiredError",
    "TransactionError",
    "TransactionConflictError",
    "TransactionSerializationError",
    "TransactionDeadlockError",
    "WatchError",
    "ConfigurationError",
    "AccessError",
    "AuthenticationError",
    "AvailabilityError",
    "BackendUnavailableError",
    "ServerOfflineError",
    "BackendError",
    "UnsupportedBackendFeatureError",
    "LogMessage",
    "WarningMessage",
    "ClientError",
    "ClientConnectionError",
    "ClientConnectionFailedError",
    "ClientConnectionFailedTemporarilyError",
    "ClientConnectionTimeoutError",
    "ClientConnectionClosedError",
    "InterfaceError",
    "QueryArgumentError",
    "MissingArgumentError",
    "UnknownArgumentError",
    "InvalidArgumentError",
    "NoDataError",
    "InternalClientError",
])
# </ERRORS-AUTOGEN>



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import enum


class EnumValue(enum.Enum):

    def __str__(self):
        return self._value_

    def __repr__(self):
        return f'<gel.EnumValue {self._value_!r}>'

    @classmethod
    def _try_from(cls, value):
        if isinstance(value, EnumValue):
            return value
        elif isinstance(value, enum.Enum):
            return cls(value.value)
        else:
            raise TypeError

    def __lt__(self, other):
        other = self._try_from(other)
        if self.__tid__ != other.__tid__:
            return NotImplemented
        return self._index_ < other._index_

    def __gt__(self, other):
        other = self._try_from(other)
        if self.__tid__ != other.__tid__:
            return NotImplemented
        return self._index_ > other._index_

    def __le__(self, other):
        other = self._try_from(other)
        if self.__tid__ != other.__tid__:
            return NotImplemented
        return self._index_ <= other._index_

    def __ge__(self, other):
        other = self._try_from(other)
        if self.__tid__ != other.__tid__:
            return NotImplemented
        return self._index_ >= other._index_

    def __eq__(self, other):
        other = self._try_from(other)
        return self is other

    def __hash__(self):
        return hash(self._value_)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2022-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import contextlib
import datetime
import queue
import socket
import ssl
import threading
import time
import typing

from . import abstract
from . import base_client
from . import con_utils
from . import errors
from . import transaction
from .protocol import blocking_proto
from .protocol.protocol import InputLanguage, OutputFormat


DEFAULT_PING_BEFORE_IDLE_TIMEOUT = datetime.timedelta(seconds=5)
MINIMUM_PING_WAIT_TIME = datetime.timedelta(seconds=1)


class BlockingIOConnection(base_client.BaseConnection):
    __slots__ = ("_ping_wait_time",)

    async def connect_addr(self, addr, timeout):
        deadline = time.monotonic() + timeout

        if isinstance(addr, str):
            # UNIX socket
            res_list = [(socket.AF_UNIX, socket.SOCK_STREAM, -1, None, addr)]
        else:
            host, port = addr
            try:
                # getaddrinfo() doesn't take timeout!!
                res_list = socket.getaddrinfo(
                    host, port, socket.AF_UNSPEC, socket.SOCK_STREAM
                )
            except socket.gaierror as e:
                # All name resolution errors are considered temporary
                err = errors.ClientConnectionFailedTemporarilyError(str(e))
                raise err from e

        for i, res in enumerate(res_list):
            af, socktype, proto, _, sa = res
            try:
                sock = socket.socket(af, socktype, proto)
            except OSError as e:
                sock.close()
                if i < len(res_list) - 1:
                    continue
                else:
                    raise con_utils.wrap_error(e) from e
            try:
                await self._connect_addr(sock, addr, sa, deadline)
            except TimeoutError:
                raise
            except Exception:
                if i < len(res_list) - 1:
                    continue
                else:
                    raise
            else:
                break

    async def _connect_addr(self, sock, addr, sa, deadline):
        try:
            time_left = deadline - time.monotonic()
            if time_left <= 0:
                raise TimeoutError
            try:
                sock.settimeout(time_left)
                sock.connect(sa)
            except OSError as e:
                raise con_utils.wrap_error(e) from e

            if not isinstance(addr, str):
                time_left = deadline - time.monotonic()
                if time_left <= 0:
                    raise TimeoutError
                try:
                    # Upgrade to TLS
                    sock.settimeout(time_left)
                    try:
                        sock = self._params.ssl_ctx.wrap_socket(
                            sock,
                            server_hostname=(
                                self._params.tls_server_name or addr[0]
                            ),
                        )
                    except ssl.CertificateError as e:
                        raise con_utils.wrap_error(e) from e
                    except ssl.SSLError as e:
                        raise con_utils.wrap_error(e) from e
                    else:
                        con_utils.check_alpn_protocol(sock)
                except OSError as e:
                    raise con_utils.wrap_error(e) from e

            time_left = deadline - time.monotonic()
            if time_left <= 0:
                raise TimeoutError

            if not isinstance(addr, str):
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

            proto = blocking_proto.BlockingIOProtocol(self._params, sock)
            proto.set_connection(self)

            try:
                await proto.wait_for(proto.connect(), time_left)
            except TimeoutError:
                raise
            except OSError as e:
                raise con_utils.wrap_error(e) from e

            self._protocol = proto
            self._addr = addr
            self._ping_wait_time = max(
                (
                    self.get_settings()
                    .get("system_config")
                    .session_idle_timeout
                    - DEFAULT_PING_BEFORE_IDLE_TIMEOUT
                ),
                MINIMUM_PING_WAIT_TIME,
            ).total_seconds()

        except Exception:
            sock.close()
            raise

    async def sleep(self, seconds):
        time.sleep(seconds)

    def is_closed(self):
        proto = self._protocol
        return not (proto and proto.sock is not None and
                    proto.sock.fileno() >= 0 and proto.connected)

    async def close(self, timeout=None):
        """Send graceful termination message wait for connection to drop."""
        if not self.is_closed():
            try:
                self._protocol.terminate()
                if timeout is None:
                    await self._protocol.wait_for_disconnect()
                else:
                    await self._protocol.wait_for(
                        self._protocol.wait_for_disconnect(), timeout
                    )
            except TimeoutError:
                self.terminate()
                raise errors.QueryTimeoutError()
            except Exception:
                self.terminate()
                raise
            finally:
                self._cleanup()

    def _dispatch_log_message(self, msg):
        for cb in self._log_listeners:
            cb(self, msg)

    async def raw_query(self, query_context: abstract.QueryContext):
        try:
            if (
                time.monotonic() - self._protocol.last_active_timestamp
                > self._ping_wait_time
            ):
                await self._protocol.ping()
        except (errors.IdleSessionTimeoutError, errors.ClientConnectionError):
            await self.connect()

        return await super().raw_query(query_context)


class _PoolConnectionHolder(base_client.PoolConnectionHolder):
    __slots__ = ()
    _event_class = threading.Event

    async def close(self, *, wait=True, timeout=None):
        if self._con is None:
            return
        await self._con.close(timeout=timeout)

    async def wait_until_released(self, timeout=None):
        return self._release_event.wait(timeout)


class _PoolImpl(base_client.BasePoolImpl):
    _holder_class = _PoolConnectionHolder

    def __init__(
        self,
        connect_args,
        *,
        max_concurrency: typing.Optional[int],
        connection_class,
    ):
        if not issubclass(connection_class, BlockingIOConnection):
            raise TypeError(
                f'connection_class is expected to be a subclass of '
                f'gel.blocking_client.BlockingIOConnection, '
                f'got {connection_class}')
        super().__init__(
            connect_args,
            connection_class,
            max_concurrency=max_concurrency,
        )

    def _ensure_initialized(self):
        if self._queue is None:
            self._queue = queue.LifoQueue(maxsize=self._max_concurrency)
            self._first_connect_lock = threading.Lock()
            self._resize_holder_pool()

    def _set_queue_maxsize(self, maxsize):
        with self._queue.mutex:
            self._queue.maxsize = maxsize

    async def _maybe_get_first_connection(self):
        with self._first_connect_lock:
            if self._working_addr is None:
                return await self._get_first_connection()

    async def acquire(self, timeout=None):
        self._ensure_initialized()

        if self._closing:
            raise errors.InterfaceError('pool is closing')

        ch = self._queue.get(timeout=timeout)
        try:
            con = await ch.acquire()
        except Exception:
            self._queue.put_nowait(ch)
            raise
        else:
            # Record the timeout, as we will apply it by default
            # in release().
            ch._timeout = timeout
            return con

    async def _release(self, holder):
        if not isinstance(holder._con, BlockingIOConnection):
            raise errors.InterfaceError(
                f'release() received invalid connection: '
                f'{holder._con!r} does not belong to any connection pool'
            )

        timeout = None
        return await holder.release(timeout)

    async def close(self, timeout=None):
        if self._closed:
            return
        self._closing = True
        try:
            if timeout is None:
                for ch in self._holders:
                    await ch.wait_until_released()
                for ch in self._holders:
                    await ch.close()
            else:
                deadline = time.monotonic() + timeout
                for ch in self._holders:
                    secs = deadline - time.monotonic()
                    if secs <= 0:
                        raise TimeoutError
                    if not await ch.wait_until_released(secs):
                        raise TimeoutError
                for ch in self._holders:
                    secs = deadline - time.monotonic()
                    if secs <= 0:
                        raise TimeoutError
                    await ch.close(timeout=secs)
        except TimeoutError as e:
            self.terminate()
            raise errors.InterfaceError(
                "client is not fully closed in {} seconds; "
                "terminating now.".format(timeout)
            ) from e
        except Exception:
            self.terminate()
            raise
        finally:
            self._closed = True
            self._closing = False


class Iteration(transaction.BaseTransaction, abstract.Executor):

    __slots__ = ("_managed", "_lock")

    def __init__(self, retry, client, iteration):
        super().__init__(retry, client, iteration)
        self._managed = False
        self._lock = threading.Lock()

    def __enter__(self):
        with self._exclusive():
            if self._managed:
                raise errors.InterfaceError(
                    'cannot enter context: already in a `with` block')
            self._managed = True
            return self

    def __exit__(self, extype, ex, tb):
        with self._exclusive():
            self._managed = False
            return self._client._iter_coroutine(self._exit(extype, ex))

    async def _ensure_transaction(self):
        if not self._managed:
            raise errors.InterfaceError(
                "Only managed retriable transactions are supported. "
                "Use `with transaction:`"
            )
        await super()._ensure_transaction()

    def _query(self, query_context: abstract.QueryContext):
        with self._exclusive():
            return self._client._iter_coroutine(super()._query(query_context))

    def _execute(self, execute_context: abstract.ExecuteContext) -> None:
        with self._exclusive():
            self._client._iter_coroutine(super()._execute(execute_context))

    @contextlib.contextmanager
    def _exclusive(self):
        if not self._lock.acquire(blocking=False):
            raise errors.InterfaceError(
                "concurrent queries within the same transaction "
                "are not allowed"
            )
        try:
            yield
        finally:
            self._lock.release()


class Retry(transaction.BaseRetry):

    def __iter__(self):
        return self

    def __next__(self):
        # Note: when changing this code consider also
        # updating AsyncIORetry.__anext__.
        if self._done:
            raise StopIteration
        if self._next_backoff:
            time.sleep(self._next_backoff)
        self._done = True
        iteration = Iteration(self, self._owner, self._iteration)
        self._iteration += 1
        return iteration


class Client(base_client.BaseClient, abstract.Executor):
    """A lazy connection pool.

    A Client can be used to manage a set of connections to the database.
    Connections are first acquired from the pool, then used, and then released
    back to the pool.  Once a connection is released, it's reset to close all
    open cursors and other resources *except* prepared statements.

    Clients are created by calling
    :func:`~gel.blocking_client.create_client`.
    """

    __slots__ = ()
    _impl_class = _PoolImpl

    def _iter_coroutine(self, coro):
        try:
            coro.send(None)
        except StopIteration as ex:
            return ex.value
        finally:
            coro.close()

    def _query(self, query_context: abstract.QueryContext):
        return self._iter_coroutine(super()._query(query_context))

    def _execute(self, execute_context: abstract.ExecuteContext) -> None:
        self._iter_coroutine(super()._execute(execute_context))

    def ensure_connected(self):
        self._iter_coroutine(self._impl.ensure_connected())
        return self

    def transaction(self) -> Retry:
        return Retry(self)

    def close(self, timeout=None):
        """Attempt to gracefully close all connections in the client.

        Wait until all pool connections are released, close them and
        shut down the pool.  If any error (including cancellation) occurs
        in ``close()`` the pool will terminate by calling
        Client.terminate() .
        """
        self._iter_coroutine(self._impl.close(timeout))

    def __enter__(self):
        return self.ensure_connected()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _describe_query(
        self,
        query: str,
        *,
        inject_type_names: bool = False,
        input_language: InputLanguage = InputLanguage.EDGEQL,
        output_format: OutputFormat = OutputFormat.BINARY,
        expect_one: bool = False,
    ) -> abstract.DescribeResult:
        return self._iter_coroutine(self._describe(abstract.DescribeContext(
            query=query,
            state=self._get_state(),
            inject_type_names=inject_type_names,
            input_language=input_language,
            output_format=output_format,
            expect_one=expect_one,
        )))


def create_client(
    dsn=None,
    *,
    max_concurrency=None,
    host: str = None,
    port: int = None,
    credentials: str = None,
    credentials_file: str = None,
    user: str = None,
    password: str = None,
    secret_key: str = None,
    database: str = None,
    branch: str = None,
    tls_ca: str = None,
    tls_ca_file: str = None,
    tls_security: str = None,
    wait_until_available: int = 30,
    timeout: int = 10,
):
    return Client(
        connection_class=BlockingIOConnection,
        max_concurrency=max_concurrency,

        # connect arguments
        dsn=dsn,
        host=host,
        port=port,
        credentials=credentials,
        credentials_file=credentials_file,
        user=user,
        password=password,
        secret_key=secret_key,
        database=database,
        branch=branch,
        tls_ca=tls_ca,
        tls_ca_file=tls_ca_file,
        tls_security=tls_security,
        wait_until_available=wait_until_available,
        timeout=timeout,
    )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import io
import os
import traceback
import unicodedata
import warnings

__all__ = (
    'EdgeDBError', 'EdgeDBMessage',
)


class Meta(type):

    def __new__(mcls, name, bases, dct):
        cls = super().__new__(mcls, name, bases, dct)

        code = dct.get('_code')
        if code is not None:
            mcls._index[code] = cls

            # If it's a base class add it to the base class index
            b1, b2, b3, b4 = _decode(code)
            if b1 == 0 or b2 == 0 or b3 == 0 or b4 == 0:
                mcls._base_class_index[(b1, b2, b3, b4)] = cls

        return cls


class EdgeDBMessageMeta(Meta):

    _base_class_index = {}
    _index = {}


class EdgeDBMessage(Warning, metaclass=EdgeDBMessageMeta):

    _code = None

    def __init__(self, severity, message):
        super().__init__(message)
        self._severity = severity

    def get_severity(self):
        return self._severity

    def get_severity_name(self):
        return _severity_name(self._severity)

    def get_code(self):
        return self._code

    @staticmethod
    def _from_code(code, severity, message, *args, **kwargs):
        cls = _lookup_message_cls(code)
        exc = cls(severity, message, *args, **kwargs)
        exc._code = code
        return exc


class EdgeDBErrorMeta(Meta):

    _base_class_index = {}
    _index = {}


class EdgeDBError(Exception, metaclass=EdgeDBErrorMeta):

    _code = None
    _query = None
    tags = frozenset()

    def __init__(self, *args, **kwargs):
        self._attrs = {}
        super().__init__(*args, **kwargs)

    def has_tag(self, tag):
        return tag in self.tags

    @property
    def _position(self):
        # not a stable API method
        return int(self._read_str_field(FIELD_POSITION_START, -1))

    @property
    def _position_start(self):
        # not a stable API method
        return int(self._read_str_field(FIELD_CHARACTER_START, -1))

    @property
    def _position_end(self):
        # not a stable API method
        return int(self._read_str_field(FIELD_CHARACTER_END, -1))

    @property
    def _line(self):
        # not a stable API method
        return int(self._read_str_field(FIELD_LINE_START, -1))

    @property
    def _col(self):
        # not a stable API method
        return int(self._read_str_field(FIELD_COLUMN_START, -1))

    @property
    def _hint(self):
        # not a stable API method
        return self._read_str_field(FIELD_HINT)

    @property
    def _details(self):
        # not a stable API method
        return self._read_str_field(FIELD_DETAILS)

    def _read_str_field(self, key, default=None):
        val = self._attrs.get(key)
        if isinstance(val, bytes):
            return val.decode('utf-8')
        elif val is not None:
            return val
        return default

    def get_code(self):
        return self._code

    def get_server_context(self):
        return self._read_str_field(FIELD_SERVER_TRACEBACK)

    @staticmethod
    def _from_code(code, *args, **kwargs):
        cls = _lookup_error_cls(code)
        exc = cls(*args, **kwargs)
        exc._code = code
        return exc

    @staticmethod
    def _from_json(data):
        exc = EdgeDBError._from_code(data['code'], data['message'])
        exc._attrs = {
            field: data[name]
            for name, field in _JSON_FIELDS.items()
            if name in data
        }
        return exc

    def __str__(self):
        msg = super().__str__()
        if SHOW_HINT and self._query and self._position_start >= 0:
            try:
                return _format_error(
                    msg,
                    self._query,
                    self._position_start,
                    max(1, self._position_end - self._position_start),
                    self._line if self._line > 0 else "?",
                    self._col if self._col > 0 else "?",
                    self._hint or "error",
                    self._details,
                )
            except Exception:
                return "".join(
                    (
                        msg,
                        LINESEP,
                        LINESEP,
                        "During formatting of the above exception, "
                        "another exception occurred:",
                        LINESEP,
                        LINESEP,
                        traceback.format_exc(),
                    )
                )
        else:
            return msg


def _lookup_cls(code: int, *, meta: type, default: type):
    try:
        return meta._index[code]
    except KeyError:
        pass

    b1, b2, b3, _ = _decode(code)

    try:
        return meta._base_class_index[(b1, b2, b3, 0)]
    except KeyError:
        pass
    try:
        return meta._base_class_index[(b1, b2, 0, 0)]
    except KeyError:
        pass
    try:
        return meta._base_class_index[(b1, 0, 0, 0)]
    except KeyError:
        pass

    return default


def _lookup_error_cls(code: int):
    return _lookup_cls(code, meta=EdgeDBErrorMeta, default=EdgeDBError)


def _lookup_message_cls(code: int):
    return _lookup_cls(code, meta=EdgeDBMessageMeta, default=EdgeDBMessage)


def _decode(code: int):
    return tuple(code.to_bytes(4, 'big'))


def _severity_name(severity):
    if severity <= EDGE_SEVERITY_DEBUG:
        return 'DEBUG'
    if severity <= EDGE_SEVERITY_INFO:
        return 'INFO'
    if severity <= EDGE_SEVERITY_NOTICE:
        return 'NOTICE'
    if severity <= EDGE_SEVERITY_WARNING:
        return 'WARNING'
    if severity <= EDGE_SEVERITY_ERROR:
        return 'ERROR'
    if severity <= EDGE_SEVERITY_FATAL:
        return 'FATAL'
    return 'PANIC'


def _format_error(msg, query, start, offset, line, col, hint, details):
    c = get_color()
    rv = io.StringIO()
    rv.write(f"{c.BOLD}{msg}{c.ENDC}{LINESEP}")
    lines = query.splitlines(keepends=True)
    num_len = len(str(len(lines)))
    rv.write(f"{c.BLUE}{'':>{num_len}} ┌─{c.ENDC} query:{line}:{col}{LINESEP}")
    rv.write(f"{c.BLUE}{'':>{num_len}} │ {c.ENDC}{LINESEP}")
    for num, line in enumerate(lines):
        length = len(line)
        line = line.rstrip()  # we'll use our own line separator
        if start >= length:
            # skip lines before the error
            start -= length
            continue

        if start >= 0:
            # Error starts in current line, write the line before the error
            first_half = repr(line[:start])[1:-1]
            line = line[start:]
            length -= start
            rv.write(f"{c.BLUE}{num + 1:>{num_len}} │   {c.ENDC}{first_half}")
            start = _unicode_width(first_half)
        else:
            # Multi-line error continues
            rv.write(f"{c.BLUE}{num + 1:>{num_len}} │ {c.FAIL}│ {c.ENDC}")

        if offset > length:
            # Error is ending beyond current line
            line = repr(line)[1:-1]
            rv.write(f"{c.FAIL}{line}{c.ENDC}{LINESEP}")
            if start >= 0:
                # Multi-line error starts
                rv.write(f"{c.BLUE}{'':>{num_len}} │ "
                         f"{c.FAIL}╭─{'─' * start}^{c.ENDC}{LINESEP}")
            offset -= length
            start = -1  # mark multi-line
        else:
            # Error is ending within current line
            first_half = repr(line[:offset])[1:-1]
            line = repr(line[offset:])[1:-1]
            rv.write(f"{c.FAIL}{first_half}{c.ENDC}{line}{LINESEP}")
            size = _unicode_width(first_half)
            if start >= 0:
                # Mark single-line error
                rv.write(f"{c.BLUE}{'':>{num_len}} │   {' ' * start}"
                         f"{c.FAIL}{'^' * size} {hint}{c.ENDC}")
            else:
                # End of multi-line error
                rv.write(f"{c.BLUE}{'':>{num_len}} │ "
                         f"{c.FAIL}╰─{'─' * (size - 1)}^ {hint}{c.ENDC}")
            break

    if details:
        rv.write(f"{LINESEP}Details: {details}")

    return rv.getvalue()


def _unicode_width(text):
    return sum(0 if unicodedata.category(c) in ('Mn', 'Cf') else
               2 if unicodedata.east_asian_width(c) == "W" else 1
               for c in text)


FIELD_HINT = 0x_00_01
FIELD_DETAILS = 0x_00_02
FIELD_SERVER_TRACEBACK = 0x_01_01

# XXX: Subject to be changed/deprecated.
FIELD_POSITION_START = 0x_FF_F1
FIELD_POSITION_END = 0x_FF_F2
FIELD_LINE_START = 0x_FF_F3
FIELD_COLUMN_START = 0x_FF_F4
FIELD_UTF16_COLUMN_START = 0x_FF_F5
FIELD_LINE_END = 0x_FF_F6
FIELD_COLUMN_END = 0x_FF_F7
FIELD_UTF16_COLUMN_END = 0x_FF_F8
FIELD_CHARACTER_START = 0x_FF_F9
FIELD_CHARACTER_END = 0x_FF_FA


EDGE_SEVERITY_DEBUG = 20
EDGE_SEVERITY_INFO = 40
EDGE_SEVERITY_NOTICE = 60
EDGE_SEVERITY_WARNING = 80
EDGE_SEVERITY_ERROR = 120
EDGE_SEVERITY_FATAL = 200
EDGE_SEVERITY_PANIC = 255


# Fields to include in the json dump of the type
_JSON_FIELDS = {
    'hint': FIELD_HINT,
    'details': FIELD_DETAILS,
    'start': FIELD_CHARACTER_START,
    'end': FIELD_CHARACTER_END,
    'line': FIELD_LINE_START,
    'col': FIELD_COLUMN_START,
}


LINESEP = os.linesep

try:
    SHOW_HINT = {"default": True, "enabled": True, "disabled": False}[
        os.getenv("EDGEDB_ERROR_HINT", "default")
    ]
except KeyError:
    warnings.warn(
        "EDGEDB_ERROR_HINT can only be one of: default, enabled or disabled",
        stacklevel=1,
    )
    SHOW_HINT = False


from gel.color import get_color



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2024-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import argparse
import warnings

import gel

from gel.codegen.generator import _get_conn_args
from .introspection import get_schema_json, GelORMWarning
from .sqla import ModelGenerator as SQLAModGen
from .sqlmodel import ModelGenerator as SQLModGen
from .django.generator import ModelGenerator as DjangoModGen


class ArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        self.exit(
            2,
            f"error: {message:s}\n",
        )


parser = ArgumentParser(
    description="Generate Python ORM code for accessing a Gel database."
)
parser.add_argument(
    "orm",
    choices=['sqlalchemy', 'sqlmodel', 'django'],
    help="Pick which ORM to generate models for.",
)
parser.add_argument("--dsn")
parser.add_argument("--credentials-file", metavar="PATH")
parser.add_argument("-I", "--instance", metavar="NAME")
parser.add_argument("-H", "--host")
parser.add_argument("-P", "--port")
parser.add_argument("-d", "--database", metavar="NAME")
parser.add_argument("-u", "--user")
parser.add_argument("--password")
parser.add_argument("--password-from-stdin", action="store_true")
parser.add_argument("--tls-ca-file", metavar="PATH")
parser.add_argument(
    "--tls-security",
    choices=["default", "strict", "no_host_verification", "insecure"],
)
parser.add_argument(
    "--out",
    help="The output directory for the generated files.",
    required=True,
)
parser.add_argument(
    "--mod",
    help="The fullname of the Python module corresponding to the output "
         "directory.",
)


def main():
    args = parser.parse_args()
    # setup client
    client = gel.create_client(**_get_conn_args(args))

    with warnings.catch_warnings(record=True) as wlist:
        warnings.simplefilter("always", GelORMWarning)

        spec = get_schema_json(client)
        generate_models(args, spec)

        for w in wlist:
            print(w.message)


def generate_models(args, spec):
    match args.orm:
        case 'sqlalchemy':
            if args.mod is None:
                parser.error('sqlalchemy requires to specify --mod')

            gen = SQLAModGen(
                outdir=args.out,
                basemodule=args.mod,
            )
            gen.render_models(spec)

        case 'sqlmodel':
            if args.mod is None:
                parser.error('sqlmodel requires to specify --mod')

            gen = SQLModGen(
                outdir=args.out,
                basemodule=args.mod,
            )
            gen.render_models(spec)

        case 'django':
            gen = DjangoModGen(
                out=args.out,
            )
            gen.render_models(spec)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import asyncio
import atexit
import contextlib
import functools
import importlib.util
import inspect
import json
import logging
import pathlib
import os
import re
import subprocess
import sys
import tempfile
import time
import unittest
import warnings

import gel
from gel import asyncio_client
from gel import blocking_client
from gel.orm.introspection import get_schema_json, GelORMWarning
from gel.orm.sqla import ModelGenerator as SQLAModGen
from gel.orm.sqlmodel import ModelGenerator as SQLModGen
from gel.orm.django.generator import ModelGenerator as DjangoModGen


log = logging.getLogger(__name__)


@contextlib.contextmanager
def silence_asyncio_long_exec_warning():
    def flt(log_record):
        msg = log_record.getMessage()
        return not msg.startswith('Executing ')

    logger = logging.getLogger('asyncio')
    logger.addFilter(flt)
    try:
        yield
    finally:
        logger.removeFilter(flt)


def _get_wsl_path(win_path):
    return (
        re.sub(r'^([A-Z]):', lambda m: f'/mnt/{m.group(1)}', win_path)
        .replace("\\", '/')
        .lower()
    )


_default_cluster = None


def _start_cluster(*, cleanup_atexit=True):
    global _default_cluster

    if isinstance(_default_cluster, Exception):
        # when starting a server fails
        # don't retry starting one for every TestCase
        # because repeating the failure can take a while
        raise _default_cluster

    if _default_cluster:
        return _default_cluster

    try:
        tmpdir = tempfile.TemporaryDirectory()
        status_file = os.path.join(tmpdir.name, 'server-status')

        # if running on windows adjust the path for WSL
        status_file_unix = _get_wsl_path(status_file)

        env = os.environ.copy()
        # Make sure the PYTHONPATH of _this_ process does
        # not interfere with the server's.
        env.pop('PYTHONPATH', None)

        gel_server = env.get('EDGEDB_SERVER_BINARY', 'edgedb-server')

        version_args = [gel_server, '--version']
        if sys.platform == 'win32':
            version_args = ['wsl', '-u', 'edgedb'] + version_args
        version_res = subprocess.run(
            version_args,
            capture_output=True,
            text=True,
        )
        is_gel = version_res.stdout.startswith('gel-server,')

        version_line = version_res.stdout
        is_gel = version_line.startswith('gel-server,')

        # The default role became admin in nightly build 9024 for 6.0
        if is_gel:
            if '6.0-dev' in version_line:
                rev = int(version_line.split('.')[2].split('+')[0])
                has_admin = rev >= 9024
            else:
                has_admin = True
        else:
            has_admin = False

        role = 'admin' if has_admin else 'edgedb'
        args = [
            gel_server,
            "--temp-dir",
            "--testmode",
            f"--emit-server-status={status_file_unix}",
            "--port=auto",
            "--auto-shutdown",
            f"--bootstrap-command=ALTER ROLE {role} {{SET password := 'test'}}",
        ]

        help_args = [gel_server, "--help"]
        if sys.platform == 'win32':
            help_args = ['wsl', '-u', 'edgedb'] + help_args

        supported_opts = subprocess.run(
            help_args,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf-8",
            env=env,
            cwd=tmpdir.name,
        ).stdout

        if "--tls-cert-mode" in supported_opts:
            args.append("--tls-cert-mode=generate_self_signed")
        elif "--generate-self-signed-cert" in supported_opts:
            args.append("--generate-self-signed-cert")

        if sys.platform == 'win32':
            args = ['wsl', '-u', 'edgedb'] + args

        if env.get('EDGEDB_DEBUG_SERVER'):
            server_stdout = None
        else:
            server_stdout = subprocess.DEVNULL

        p = subprocess.Popen(
            args,
            env=env,
            cwd=tmpdir.name,
            stdout=server_stdout,
            stderr=subprocess.STDOUT,
        )

        for _ in range(600):
            if p.poll() is not None:
                raise RuntimeError(
                    'Database server crashed before signalling READY status.'
                    ' Run with `env EDGEDB_DEBUG_SERVER=1` to debug.')
            try:
                with open(status_file, 'rb') as f:
                    for line in f:
                        if line.startswith(b'READY='):
                            break
                    else:
                        raise RuntimeError('not ready')
                break
            except Exception:
                time.sleep(1)
        else:
            raise RuntimeError('server status file not found')

        data = json.loads(line.split(b'READY=')[1])
        con_args = dict(host='localhost', port=data['port'])
        if 'tls_cert_file' in data:
            if sys.platform == 'win32':
                con_args['tls_ca_file'] = os.path.join(
                    tmpdir.name, "edbtlscert.pem"
                )
                subprocess.check_call(
                    [
                        'wsl',
                        '-u',
                        'edgedb',
                        'cp',
                        data['tls_cert_file'],
                        _get_wsl_path(con_args['tls_ca_file'])
                    ]
                )
            else:
                con_args['tls_ca_file'] = data['tls_cert_file']

        client = gel.create_client(password='test', **con_args)
        client.ensure_connected()
        client.execute("""
            # Set session_idle_transaction_timeout to 5 minutes.
            CONFIGURE INSTANCE SET session_idle_transaction_timeout :=
                <duration>'5 minutes';
        """)
        _default_cluster = {
            'proc': p,
            'client': client,
            'con_args': con_args,
        }

        if 'tls_cert_file' in data:
            # Keep the temp dir which we also copied the cert from WSL
            _default_cluster['_tmpdir'] = tmpdir

        atexit.register(client.close)
    except Exception as e:
        _default_cluster = e
        raise e

    return _default_cluster


class TestCaseMeta(type(unittest.TestCase)):
    _database_names = set()

    @staticmethod
    def _iter_methods(bases, ns):
        for base in bases:
            for methname in dir(base):
                if not methname.startswith('test_'):
                    continue

                meth = getattr(base, methname)
                if not inspect.iscoroutinefunction(meth):
                    continue

                yield methname, meth

        for methname, meth in ns.items():
            if not methname.startswith('test_'):
                continue

            if not inspect.iscoroutinefunction(meth):
                continue

            yield methname, meth

    @classmethod
    def wrap(mcls, meth):
        @functools.wraps(meth)
        def wrapper(self, *args, __meth__=meth, **kwargs):
            try_no = 1

            while True:
                try:
                    # There might be unobvious serializability
                    # anomalies across the test suite, so, rather
                    # than hunting them down every time, simply
                    # retry the test.
                    self.loop.run_until_complete(
                        __meth__(self, *args, **kwargs))
                except gel.TransactionSerializationError:
                    if try_no == 3:
                        raise
                    else:
                        self.loop.run_until_complete(self.client.execute(
                            'ROLLBACK;'
                        ))
                        try_no += 1
                else:
                    break

        return wrapper

    @classmethod
    def add_method(mcls, methname, ns, meth):
        ns[methname] = mcls.wrap(meth)

    def __new__(mcls, name, bases, ns):
        for methname, meth in mcls._iter_methods(bases, ns.copy()):
            if methname in ns:
                del ns[methname]
            mcls.add_method(methname, ns, meth)

        cls = super().__new__(mcls, name, bases, ns)
        if not ns.get('BASE_TEST_CLASS') and hasattr(cls, 'get_database_name'):
            dbname = cls.get_database_name()

            if name in mcls._database_names:
                raise TypeError(
                    f'{name} wants duplicate database name: {dbname}')

            mcls._database_names.add(name)

        return cls


class TestCase(unittest.TestCase, metaclass=TestCaseMeta):
    @classmethod
    def setUpClass(cls):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        cls.loop = loop

    @classmethod
    def tearDownClass(cls):
        cls.loop.close()
        asyncio.set_event_loop(None)

    def add_fail_notes(self, **kwargs):
        if not hasattr(self, 'fail_notes'):
            self.fail_notes = {}
        self.fail_notes.update(kwargs)

    @contextlib.contextmanager
    def annotate(self, **kwargs):
        # Annotate the test in case the nested block of code fails.
        try:
            yield
        except Exception:
            self.add_fail_notes(**kwargs)
            raise

    @contextlib.contextmanager
    def assertRaisesRegex(self, exception, regex, msg=None,
                          **kwargs):
        with super().assertRaisesRegex(exception, regex, msg=msg):
            try:
                yield
            except BaseException as e:
                if isinstance(e, exception):
                    for attr_name, expected_val in kwargs.items():
                        val = getattr(e, attr_name)
                        if val != expected_val:
                            raise self.failureException(
                                f'{exception.__name__} context attribute '
                                f'{attr_name!r} is {val} (expected '
                                f'{expected_val!r})') from e
                raise

    def addCleanup(self, func, *args, **kwargs):
        @functools.wraps(func)
        def cleanup():
            res = func(*args, **kwargs)
            if inspect.isawaitable(res):
                self.loop.run_until_complete(res)
        super().addCleanup(cleanup)


class ClusterTestCase(TestCase):

    BASE_TEST_CLASS = True

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.cluster = _start_cluster(cleanup_atexit=True)


class TestAsyncIOClient(gel.AsyncIOClient):
    def _clear_codecs_cache(self):
        self._impl.codecs_registry.clear_cache()

    @property
    def connection(self):
        return self._impl._holders[0]._con

    @property
    def dbname(self):
        return self._impl._working_params.database

    @property
    def is_proto_lt_1_0(self):
        return self.connection._protocol.is_legacy


class TestClient(gel.Client):
    @property
    def connection(self):
        return self._impl._holders[0]._con

    @property
    def is_proto_lt_1_0(self):
        return self.connection._protocol.is_legacy

    @property
    def dbname(self):
        return self._impl._working_params.database


class ConnectedTestCaseMixin:
    is_client_async = True

    @classmethod
    def make_test_client(
        cls, *,
        cluster=None,
        database='edgedb',
        user='edgedb',
        password='test',
        host=...,
        port=...,
        connection_class=...,
    ):
        conargs = cls.get_connect_args(
            cluster=cluster, database=database, user=user, password=password)
        if host is not ...:
            conargs['host'] = host
        if port is not ...:
            conargs['port'] = port
        if connection_class is ...:
            connection_class = (
                asyncio_client.AsyncIOConnection
                if cls.is_client_async
                else blocking_client.BlockingIOConnection
            )
        return (TestAsyncIOClient if cls.is_client_async else TestClient)(
            connection_class=connection_class,
            max_concurrency=1,
            **conargs,
        )

    @classmethod
    def get_connect_args(cls, *,
                         cluster=None,
                         database='edgedb',
                         user='edgedb',
                         password='test'):
        conargs = cls.cluster['con_args'].copy()
        conargs.update(dict(user=user,
                            password=password,
                            database=database))
        return conargs

    @classmethod
    def adapt_call(cls, coro):
        return cls.loop.run_until_complete(coro)


class DatabaseTestCase(ClusterTestCase, ConnectedTestCaseMixin):
    SETUP = None
    TEARDOWN = None
    SCHEMA = None
    DEFAULT_MODULE = 'test'

    SETUP_METHOD = None
    TEARDOWN_METHOD = None

    BASE_TEST_CLASS = True
    TEARDOWN_RETRY_DROP_DB = 1

    def setUp(self):
        if self.SETUP_METHOD:
            self.adapt_call(
                self.client.execute(self.SETUP_METHOD))

        super().setUp()

    def tearDown(self):
        try:
            if self.TEARDOWN_METHOD:
                self.adapt_call(
                    self.client.execute(self.TEARDOWN_METHOD))
        finally:
            try:
                if self.client.connection.is_in_transaction():
                    raise AssertionError(
                        'test connection is still in transaction '
                        '*after* the test')

            finally:
                super().tearDown()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dbname = cls.get_database_name()

        cls.admin_client = None

        class_set_up = os.environ.get('EDGEDB_TEST_CASES_SET_UP')

        # Only open an extra admin connection if necessary.
        if not class_set_up:
            script = f'CREATE DATABASE {dbname};'
            cls.admin_client = cls.make_test_client()
            cls.adapt_call(cls.admin_client.execute(script))

        cls.client = cls.make_test_client(database=dbname)

        if not class_set_up:
            script = cls.get_setup_script()
            if script:
                # The setup is expected to contain a CREATE MIGRATION,
                # which needs to be wrapped in a transaction.
                if cls.is_client_async:
                    async def execute():
                        async for tr in cls.client.transaction():
                            async with tr:
                                await tr.execute(script)
                else:
                    def execute():
                        for tr in cls.client.transaction():
                            with tr:
                                tr.execute(script)
                cls.adapt_call(execute())

    @classmethod
    def get_database_name(cls):
        if cls.__name__.startswith('TestEdgeQL'):
            dbname = cls.__name__[len('TestEdgeQL'):]
        elif cls.__name__.startswith('Test'):
            dbname = cls.__name__[len('Test'):]
        else:
            dbname = cls.__name__

        return dbname.lower()

    @classmethod
    def get_setup_script(cls):
        script = ''
        schema = []

        # Look at all SCHEMA entries and potentially create multiple
        # modules, but always create the test module, if not `default`.
        if cls.DEFAULT_MODULE != 'default':
            schema.append(f'\nmodule {cls.DEFAULT_MODULE} {{}}')
        for name, val in cls.__dict__.items():
            m = re.match(r'^SCHEMA(?:_(\w+))?', name)
            if m:
                module_name = (
                    m.group(1) or cls.DEFAULT_MODULE
                ).lower().replace('_', '::')

                with open(val, 'r') as sf:
                    module = sf.read()

                schema.append(f'\nmodule {module_name} {{ {module} }}')

        # Don't wrap the script into a transaction here, so that
        # potentially it's easier to stitch multiple such scripts
        # together in a fashion similar to what `edb inittestdb` does.
        script += f'\nSTART MIGRATION TO {{ {"".join(schema)} }};'
        script += f'\nPOPULATE MIGRATION; \nCOMMIT MIGRATION;'

        if cls.SETUP:
            if not isinstance(cls.SETUP, (list, tuple)):
                scripts = [cls.SETUP]
            else:
                scripts = cls.SETUP

            for scr in scripts:
                if '\n' not in scr and os.path.exists(scr):
                    with open(scr, 'rt') as f:
                        setup = f.read()
                else:
                    setup = scr

                script += '\n' + setup

        return script.strip(' \n')

    @classmethod
    def tearDownClass(cls):
        script = ''

        if cls.TEARDOWN:
            script = cls.TEARDOWN.strip()

        try:
            if script:
                cls.adapt_call(
                    cls.client.execute(script))
        finally:
            try:
                if cls.is_client_async:
                    cls.adapt_call(cls.client.aclose())
                else:
                    cls.client.close()

                dbname = cls.get_database_name()
                script = f'DROP DATABASE {dbname};'

                retry = cls.TEARDOWN_RETRY_DROP_DB
                for i in range(retry):
                    try:
                        cls.adapt_call(
                            cls.admin_client.execute(script))
                    except gel.errors.ExecutionError:
                        if i < retry - 1:
                            time.sleep(0.1)
                        else:
                            raise
                    except gel.errors.UnknownDatabaseError:
                        break

            except Exception:
                log.exception('error running teardown')
                # skip the exception so that original error is shown instead
                # of finalizer error
            finally:
                try:
                    if cls.admin_client is not None:
                        if cls.is_client_async:
                            cls.adapt_call(
                                cls.admin_client.aclose())
                        else:
                            cls.admin_client.close()
                finally:
                    super().tearDownClass()


class AsyncQueryTestCase(DatabaseTestCase):
    BASE_TEST_CLASS = True


class SyncQueryTestCase(DatabaseTestCase):
    BASE_TEST_CLASS = True
    TEARDOWN_RETRY_DROP_DB = 5
    is_client_async = False

    @classmethod
    def adapt_call(cls, result):
        return result


class ORMTestCase(SyncQueryTestCase):
    MODEL_PACKAGE = None
    DEFAULT_MODULE = 'default'

    @classmethod
    def setUpClass(cls):
        # ORMs rely on psycopg2 to connect to Postgres and thus we
        # need it to run tests. Unfortunately not all test environemnts might
        # have psycopg2 installed, as long as we run this in the test
        # environments that have this, it is fine since we're not expecting
        # different functionality based on flavours of psycopg2.
        if importlib.util.find_spec("psycopg2") is None:
            raise unittest.SkipTest("need psycopg2 for ORM tests")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", GelORMWarning)

            super().setUpClass()

            class_set_up = os.environ.get('EDGEDB_TEST_CASES_SET_UP')
            if not class_set_up:
                # We'll need a temp directory to setup the generated Python
                # package
                cls.tmpormdir = tempfile.TemporaryDirectory()
                sys.path.append(cls.tmpormdir.name)
                # Now that the DB is setup, generate the ORM models from it
                cls.spec = get_schema_json(cls.client)
                cls.setupORM()

    @classmethod
    def setupORM(cls):
        raise NotImplementedError

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        # cleanup the temp modules
        sys.path.remove(cls.tmpormdir.name)
        cls.tmpormdir.cleanup()


class SQLATestCase(ORMTestCase):
    @classmethod
    def setupORM(cls):
        gen = SQLAModGen(
            outdir=os.path.join(cls.tmpormdir.name, cls.MODEL_PACKAGE),
            basemodule=cls.MODEL_PACKAGE,
        )
        gen.render_models(cls.spec)

    @classmethod
    def get_dsn_for_sqla(cls):
        cargs = cls.get_connect_args(database=cls.get_database_name())
        dsn = (
            f'postgresql://{cargs["user"]}:{cargs["password"]}'
            f'@{cargs["host"]}:{cargs["port"]}/{cargs["database"]}'
        )

        return dsn


class SQLModelTestCase(ORMTestCase):
    @classmethod
    def setupORM(cls):
        gen = SQLModGen(
            outdir=os.path.join(cls.tmpormdir.name, cls.MODEL_PACKAGE),
            basemodule=cls.MODEL_PACKAGE,
        )
        gen.render_models(cls.spec)

    @classmethod
    def get_dsn_for_sqla(cls):
        cargs = cls.get_connect_args(database=cls.get_database_name())
        dsn = (
            f'postgresql://{cargs["user"]}:{cargs["password"]}'
            f'@{cargs["host"]}:{cargs["port"]}/{cargs["database"]}'
        )

        return dsn


APPS_PY = '''\
from django.apps import AppConfig


class TestConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = {name!r}
'''

SETTINGS_PY = '''\
from pathlib import Path

mysettings = dict(
    INSTALLED_APPS=[
        '{appname}.apps.TestConfig',
        'gel.orm.django.gelmodels.apps.GelPGModel',
    ],
    DATABASES={{
        'default': {{
            'ENGINE': 'django.db.backends.postgresql',
            'NAME': {database!r},
            'USER': {user!r},
            'PASSWORD': {password!r},
            'HOST': {host!r},
            'PORT': {port!r},
        }}
    }},
)
'''


class DjangoTestCase(ORMTestCase):
    @classmethod
    def setupORM(cls):
        pkgbase = os.path.join(cls.tmpormdir.name, cls.MODEL_PACKAGE)
        # Set up the package for testing Django models
        os.mkdir(pkgbase)
        open(os.path.join(pkgbase, '__init__.py'), 'w').close()
        with open(os.path.join(pkgbase, 'apps.py'), 'wt') as f:
            print(
                APPS_PY.format(name=cls.MODEL_PACKAGE),
                file=f,
            )

        with open(os.path.join(pkgbase, 'settings.py'), 'wt') as f:
            cargs = cls.get_connect_args(database=cls.get_database_name())
            print(
                SETTINGS_PY.format(
                    appname=cls.MODEL_PACKAGE,
                    database=cargs["database"],
                    user=cargs["user"],
                    password=cargs["password"],
                    host=cargs["host"],
                    port=cargs["port"],
                ),
                file=f,
            )

        models = os.path.join(pkgbase, 'models.py')
        gen = DjangoModGen(out=models)
        gen.render_models(cls.spec)


_lock_cnt = 0


def gen_lock_key():
    global _lock_cnt
    _lock_cnt += 1
    return os.getpid() * 1000 + _lock_cnt


if os.environ.get('USE_UVLOOP'):
    import uvloop
    uvloop.install()



# Auto-generated shim
TYPE_CHECKING = False
if TYPE_CHECKING:
    from gel.pgproto.pgproto import *  # noqa
import gel.pgproto.pgproto as _mod
import sys as _sys
_cur = _sys.modules['edgedb.pgproto.pgproto']
for _k in vars(_mod):
    if not _k.startswith('__') or _k in ('__all__', '__doc__'):
        setattr(_cur, _k, getattr(_mod, _k))
del _cur
del _sys
del _mod
del _k
